"""
motifscan.motif.matrix
----------------------

Module for motif position matrix.
"""

import numpy as np

from motifscan.genome import bases


class PositionMatrix:
    """Two-dimensional generic position matrix."""

    def __init__(self, values, name=None, matrix_id=None):
        if len(values) != 4:  # check if values have exactly 4 rows
            raise ValueError("values should have exactly 4 rows for A/C/G/T")

        self.matrix = np.asarray(values)
        if self.matrix.ndim != 2:
            raise ValueError("values should have 2 dimensions in (4 x N)")
        if not (np.issubdtype(self.matrix.dtype, np.integer) or np.issubdtype(
                self.matrix.dtype, np.floating)):
            raise ValueError("values should be integers or floating numbers")

        self._length = self.matrix.shape[1]
        if self._length == 0:
            raise ValueError("values should have at least 1 position per row")

        self.name = name
        self.matrix_id = matrix_id

    @property
    def shape(self):
        """Shape of the position matrix (4 x N)."""
        return self.matrix.shape

    @property
    def length(self):
        """Length of the position matrix (N)."""
        return self._length

    def __len__(self):
        return self._length

    def __str__(self):
        return "A {}\nC {}\nG {}\nT {}\n".format(*self.matrix)


class PositionFrequencyMatrix(PositionMatrix):
    """Two-dimensional position frequency matrix (PFM).

    Parameters
    ----------
    values : array_like
        Non-negative integers which compose the 4 x N position frequency
        matrix. Rows represent the vectors of N positions in the order of
        A, C, G, T.
    name : str, optional
        The name of the PFM.
    matrix_id : str, optional
        The id of the matrix.
    """

    def __init__(self, values, name=None, matrix_id=None):
        super().__init__(values, name, matrix_id)
        if not np.issubdtype(self.matrix.dtype, np.integer) or np.any(
                self.matrix < 0):
            raise ValueError("values in PFM should be non-negative integers")
        if np.any(self.matrix.sum(axis=0) == 0):
            raise ValueError("all values of a PFM position are 0")

    def to_ppm(self, normalize=True, pseudo=0.001):
        """Convert the position frequency matrix (PFM) to the position
        probability matrix (PPM).

        Parameters
        ----------
        normalize : bool, optional
            Whether normalize the PPM or not. Default=True.
        pseudo : float, optional
            The pseudo probability given to positions which contain 0.
            Only valid when `normalize` is True. Default=0.001.
            In this case, frequencies [0, 0, 10, 10] will be normalized to
            probabilities [0.001, 0.001, 0.499, 0.499].

        Returns
        -------
        ppm : PositionProbabilityMatrix
            Converted PPM.
        """
        ppm = PositionProbabilityMatrix(
            values=self.matrix / self.matrix.sum(axis=0), name=self.name,
            matrix_id=self.matrix_id)
        if normalize:
            ppm.normalize(pseudo)
        return ppm


class PositionProbabilityMatrix(PositionMatrix):
    """Two-dimensional position probability matrix (PPM).

    Parameters
    ----------
    values : array_like
        Non-negative probabilities which compose the 4 x N position probability
        matrix. Rows represent the vectors of N positions in the order of
        A, C, G, T.
    name : str, optional
        The name of the PPM.
    matrix_id : str, optional
        The id of the matrix.
    """

    def __init__(self, values, name=None, matrix_id=None):
        super().__init__(values, name, matrix_id)
        if np.any(self.matrix < 0):
            raise ValueError("values in PPM should be non-negative numbers")
        if np.any(self.matrix.sum(axis=0) == 0):
            raise ValueError("all values of a PPM position are 0")
        if not np.allclose(self.matrix.sum(axis=0), 1):
            raise ValueError("the sum probability of a PPM position is not 1")

    def normalize(self, pseudo=0.001):
        """Normalize the PPM, assign a pseudo probability for every zero value
        and normalize the sum probability to 1 for each position.

        Parameters
        ----------
        pseudo : float, optional
            The pseudo probability given to positions which contain 0.
            Only valid when `normalize` is True. Default=0.001.
            In this case, frequencies [0, 0, 10, 10] will be normalized to
            probabilities [0.001, 0.001, 0.499, 0.499].

        Notes
        -----
        This method is guaranteed only when the PPM is valid, which means the
        sum probability for each position should equal to or very close to 1.
        """
        if not 0 < pseudo < 0.25:
            raise ValueError("the range of pseudo should be (0, 0.25)")
        pseudo_count = pseudo / (1 - 4 * pseudo)
        zero_cols = np.any(self.matrix == 0, axis=0)
        self.matrix[:, zero_cols] += pseudo_count
        self.matrix = self.matrix / self.matrix.sum(axis=0)

    def to_pwm(self, bg_freq=None):
        """Convert the position probability matrix (PPM) to the position
        weight matrix (PWM).

        Parameters
        ----------
        bg_freq : dict of {str: float}, optional
            A dict object which stores the nucleotide frequencies used to
            calculate the weight. Default=None and [0.25, 0.25, 0.25, 0.25]
            will be used for A, C, G, T.

        Returns
        -------
        pwm : PositionWeightMatrix
            Converted PWM.
        """
        if bg_freq is None:
            bg_freq = {base: 0.25 for base in bases}
        bg = np.asarray([bg_freq[base] for base in bases]).reshape(4, 1)
        pwm = PositionWeightMatrix(
            values=np.around(np.log(self.matrix / bg), 5),
            name=self.name, matrix_id=self.matrix_id)
        return pwm


class PositionWeightMatrix(PositionMatrix):
    """Two-dimensional position weight matrix (PWM).

    Parameters
    ----------
    values : array_like
        Values which compose the 4 x N position weight matrix. Rows represent
        the vectors of N positions in the order of A, C, G, T.
    name : str, optional
        The name of the PWM.
    matrix_id : str, optional
        The id of the matrix.
    cutoffs : dict of {str: float}, optional
        Score cutoffs for motif occurrences under different p values.
    """

    def __init__(self, values, name=None, matrix_id=None, cutoffs=None):
        super().__init__(values, name, matrix_id)
        self._max_raw_score = None
        self._min_raw_score = None
        self.cutoffs = cutoffs

    @property
    def max_raw_score(self):
        """Calculate the possible maximum raw score for the PWM."""
        if self._max_raw_score is None:
            self._max_raw_score = self.matrix.max(axis=0).sum()
        return self._max_raw_score

    @property
    def min_raw_score(self):
        """Calculate the possible minimum raw score for the PWM."""
        if self._min_raw_score is None:
            self._min_raw_score = self.matrix.min(axis=0).sum()
        return self._min_raw_score

    def score(self, sequence):
        """Calculate the score for certain sequence, which is the ratio of the
        raw score to the possible maximum raw score (raw_score/max_raw_score).

        Parameters
        ----------
        sequence : str
            The sequence to be scored under the PWM.

        Returns
        -------
        score : float
            Score of the specified sequence.
        """
        if len(sequence) != self.length:
            raise ValueError("sequence should have the same length as the PWM")
        row_idx = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
        raw_score = 0
        for col_idx, nt in enumerate(sequence.upper()):
            try:
                raw_score += self.matrix[row_idx[nt], col_idx]
            except KeyError:
                continue
        score = raw_score / self.max_raw_score
        return score



"""
motifscan.region
----------------

Module for MotifScan genomic regions.
"""

import logging
from motifscan.region.parsers import get_region_parser

logger = logging.getLogger(__name__)

REGION_FORMATS = ['bed', 'bed3-summit', 'macs', 'macs2', 'narrowpeak',
                  'broadpeak', 'manorm']


class GenomicRegion:
    """Class for a genomic region.

    Parameters
    ----------
    chrom : str
        The chromosome name of the region.
    start : int
        The start coordinate of the region.
    end : int
        The end coordinate of the region.
    summit : int, optional
        The summit coordinate of the region. If not specified, the middle point
        will be taken as the summit.
    score : float, optional
        The score of the region.

    Attributes
    ----------
    chrom : str
        The chromosome name of the region.
    start : int
        The start coordinate of the region.
    end : int
        The end coordinate of the region.
    summit : int
        The summit coordinate of the region.
    score : float or None
        The score of the region or None if not specified.

    Notes
    -----
    The coordinates are 0-based, which means the region range is [start, end)
    and `start` <= `summit` < `end` is strictly inspected.
    """

    def __init__(self, chrom, start, end, summit=None, score=None):
        self.chrom = chrom
        self.start = int(start)
        self.end = int(end)
        if self.start >= self.end:
            raise ValueError(
                f"expect start < end, got: start={start} end={end}")
        if summit is not None:
            self.summit = int(summit)
        else:
            self.summit = (self.start + self.end) // 2
        if not self.start <= self.summit < self.end:
            raise ValueError(
                f"expect start <= summit < end, got start={start} "
                f"summit={summit} end={end}")
        self.score = score

    def __repr__(self):
        return f"GenomicRegion({self.chrom}:{self.start}-{self.end})"


def load_motifscan_regions(path, format='bed'):
    """Load genomic regions from the specified path.

    Parameters
    ----------
    path : str
        Path to load the genomic regions.
    format : str, optional
        File format, default='bed'.

    Returns
    -------
    regions : list of `GenomicRegion`
        Loaded genomic regions.
    """
    logger.info(f"Loading genomic regions from {path} [{format}]")
    parser = get_region_parser(format)()
    regions = []
    for chrom, start, end, summit, score in parser.parse(path):
        regions.append(GenomicRegion(chrom, start, end, summit, score))
    logger.info(f"Loaded {len(regions)} genomic regions")
    return regions



import os

import pytest

from motifscan.cli.main import configure_parser_main
from motifscan.cli.motif import run
from motifscan.config import Config

# from motifscan.motif import

parser = configure_parser_main()


def test_cli_motif_list(tmp_dir, capsys):
    config_file = os.path.join(tmp_dir, "test_cli_motif.motifscanrc")
    config = Config(config_file)
    config.set_motif_path("motif_set1", "/path/to/motif1")
    config.set_motif_path("motif_set2", "/path/to/motif2")
    config.write()

    args = parser.parse_args(["motif", "--list"])
    run(args=args, config_file=config_file)
    captured = capsys.readouterr()
    assert captured.out == "motif_set1\nmotif_set2\n"


def test_cli_motif_list_remote(capsys):
    args = parser.parse_args(["motif", "--list-remote"])
    run(args=args)
    captured = capsys.readouterr()
    assert captured.out
    args = parser.parse_args(
        ["motif", "--list-remote", "--database", "jaspar_collections"])
    run(args=args)
    captured = capsys.readouterr()
    assert captured.out


def test_cli_motif_install(motif_root, tmp_dir):
    config_file = os.path.join(tmp_dir, "test_cli_motif.motifscanrc")
    config = Config(config_file)
    config.set_motif_dir(tmp_dir)
    config.write()

    pfms_path = os.path.join(motif_root, "test", "test_pfms.jaspar")
    args = parser.parse_args(
        ["motif", "--install", "-n", "test_motif", "-i", pfms_path])
    run(args=args, config_file=config_file)

    motif_path = os.path.join(tmp_dir, "test_motif")
    assert os.path.isfile(os.path.join(motif_path, "test_motif_pfms.jaspar"))

    config = Config(config_file)
    assert config.has_motif_set("test_motif")
    assert config.get_motif_path("test_motif") == motif_path


def test_cli_motif_uninstall(motif_root, tmp_dir):
    config_file = os.path.join(tmp_dir, "test_cli_motif.motifscanrc")
    config = Config(config_file)
    config.set_motif_dir(tmp_dir)
    config.write()

    args = parser.parse_args(["motif", "--uninstall", "test_motif"])
    run(args=args, config_file=config_file)

    config = Config(config_file)
    assert not config.has_motif_set("test_motif")

    motif_path = os.path.join(tmp_dir, "test_motif")
    assert not os.path.isdir(motif_path)

    args = parser.parse_args(["motif", "--uninstall", "test_motif1"])
    with pytest.raises(SystemExit):
        run(args=args, config_file=config_file)



"""
motifscan.motif
---------------

Module for motif related classes and functions.
"""

import logging
import os
import re

import numpy as np

from motifscan.config import Config
from motifscan.exceptions import PfmsFileNotFoundError, \
    PwmsFileNotFoundError, PfmsJasparFormatError, PwmsMotifScanFormatError
from motifscan.genome import bases
from motifscan.motif.matrix import PositionFrequencyMatrix, \
    PositionWeightMatrix

logger = logging.getLogger(__name__)

pfms_path_fmt = os.path.join("{0}", "{1}_pfms.jaspar")
pwms_path_fmt = os.path.join("{0}", "{1}_{2}_pwms.motifscan")


class MotifMatrices:
    """Generic class for motif matrices."""

    def __init__(self):
        self._matrices = []

    def __iter__(self):
        yield from self._matrices

    def __len__(self):
        return len(self._matrices)

    def append(self, item):
        self._matrices.append(item)

    def extend(self, items):
        self._matrices.extend(items)


class MotifPfms(MotifMatrices):
    """Class for a set of motif PFMs (Position Frequency Matrices).

    Parameters
    ----------
    pfms : list of `PositionFrequencyMatrix`, optional
        PFMs used to construct the class.
    name : str, optional
        The name of the motif PFMs set.

    Attributes
    ----------
    name : str or None
        The name of the motif PFMs set, or None if not specified.
    """

    def __init__(self, pfms=None, name=None):
        super().__init__()
        self.name = name
        if pfms is not None:
            for pfm in list(pfms):
                if isinstance(pfm, PositionFrequencyMatrix):
                    self.append(pfm)
                else:
                    raise ValueError(f"invalid PFM item: {pfm!r}")

    @staticmethod
    def _parse_jaspar_pfms(path):
        """Parse motif PFMs in JASPAR format.

        The regular expression pattern of the header/matrix line is inspired by
        `Biopython`.

        JASPAR PFM Example:

            >MA0006.1	Ahr::Arnt
            A  [     3      0      0      0      0      0 ]
            C  [     8      0     23      0      0      0 ]
            G  [     2     23      0     23      0     24 ]
            T  [    11      1      1      1     24      0 ]

        Raises
        ------
        PfmsJasparFormatError
            If the file does not strictly follow the JASPAR PFMs format.
        """
        header_pattern = re.compile(r"^>\s*(\S+)(\s+(\S+))?")
        matrix_new_pattern = re.compile(r"\s*([ACGT])\s*\[\s*(.+)\s*\]")
        matrix_old_pattern = re.compile(r"\s*(.+)\s*")

        pfms = []
        line_num = 0
        expect_header = True  # expect header (True) or matrix line (False)
        with open(path, 'r') as fin:
            for line in fin:
                line_num += 1
                line = line.strip()
                if not line:  # skip blank lines
                    continue

                m_header = header_pattern.match(line)
                m_matrix_new = matrix_new_pattern.match(line)
                m_matrix_old = matrix_old_pattern.match(line)

                if bool(m_header) != expect_header:  # format check
                    raise PfmsJasparFormatError(line_num, line)

                if m_header:
                    matrix_id = m_header.group(1)
                    name = m_header.group(3)
                    n_matrix = 0
                    values = []
                    expect_header = False
                else:
                    if m_matrix_new:
                        base = m_matrix_new.group(1)
                        if base != bases[n_matrix]:
                            raise PfmsJasparFormatError(line_num, line)
                        tmp_values = m_matrix_new.group(2).split()
                    elif m_matrix_old:
                        tmp_values = m_matrix_old.group(1).split()
                    else:  # neither a header nor a matrix line
                        raise PfmsJasparFormatError(line_num, line)
                    try:
                        values.append(list(map(int, tmp_values)))
                    except (ValueError, TypeError):
                        raise PfmsJasparFormatError(line_num, line)
                    n_matrix += 1
                    if n_matrix == 4:
                        pfm = PositionFrequencyMatrix(values=values, name=name,
                                                      matrix_id=matrix_id)
                        pfms.append(pfm)
                        expect_header = True

            if not expect_header:  # check whether the last matrix is complete
                raise PfmsJasparFormatError(line_num + 1, '')
        return pfms

    def read_pfms(self, path, format='jaspar'):
        """Read motif PFMs.

        Parameters
        ----------
        path : str
            Path to load the PFMs.
        format : {'jaspar'}, optional
            PFMs file format, default='jaspar'.
        """
        if format not in ['jaspar']:
            raise ValueError(f"invalid motif PFMs file format: {format!r}")
        logger.debug(f"Reading motif PFMs from {path} [{format}]")
        pfms = self._parse_jaspar_pfms(path)
        self.extend(pfms)
        logger.debug(f"Found {len(pfms)} motif PFMs")


class MotifPwms(MotifMatrices):
    """Class for a set of motif PWMs (Position Weight Matrices).

    Parameters
    ----------
    pwms : list of `PositionWeightMatrix`, optional
        PWMs used to construct the class.
    name : str, optional
        The name of the motif PWMs set.
    genome : str, optional
        The name of the genome assembly under which these PWMs are built.

    Attributes
    ----------
    name : str or None
        The name of the motif PWMs set, or None if not specified.
    genome : str or None
        The name of the genome assembly under which these PWMs are built, or
        None if not specified.
    """

    def __init__(self, pwms=None, name=None, genome=None):
        super().__init__()
        self.name = name
        self.genome = genome
        if pwms is not None:
            for pwm in list(pwms):
                if isinstance(pwm, PositionWeightMatrix):
                    self.append(pwm)
                else:
                    raise ValueError(f"invalid PWM item: {pwm!r}")

    def save_built_pwms(self):
        """Save built motif PWMs."""
        logger.info(
            f"Saving motif PWMs {self.name!r} under assembly {self.genome!r}")
        motif_dir = Config().get_motif_path(self.name)
        pwms_path = pwms_path_fmt.format(motif_dir, self.name, self.genome)
        self.write_motifscan_pwms(pwms_path)

    def write_motifscan_pwms(self, path):
        """Write motif PWMs in MotifScan format.

        Parameters
        ----------
        path : str
            The file path to write the MotifScan PWMs.
        """
        logger.debug(f"Writing MotifScan PWMs to {path}")
        with open(path, 'w') as f_out:
            for pwm in self:
                f_out.write(f">{pwm.matrix_id}\t{pwm.name}\tPWM\n")
                for idx, base in enumerate(bases):
                    values_str = '\t'.join(
                        map(lambda x: f'{x:8.5f}', pwm.matrix[idx]))
                    f_out.write(f"{base} [{values_str}]\n")
                for p, cutoff in pwm.cutoffs.items():
                    f_out.write(f"Cutoff_p{p}\t{cutoff}\n")

    def read_motifscan_pwms(self, path):
        """Read PWMs in MotifScan format.

         MotifScan PWM Example:

            >MA0006.1   Ahr::Arnt   PWM
            A [-0.85815 -5.68647 -5.68647 -5.68647 -5.68647 -5.68647]
            C [ 0.48657 -5.32257  1.53966 -5.32257 -5.32257 -5.32257]
            G [-0.90016  1.53922 -5.32301  1.53922 -5.32301  1.58174]
            T [ 0.43981 -1.93828 -1.93828 -1.93828  1.21696 -5.68779]
            Cutoff_p1e-3	0.55403
            Cutoff_p1e-4	0.82985
            Cutoff_p1e-5	1.0

        Parameters
        ----------
        path : str
            The file path to read the MotifScan PWMs.

        Raises
        ------
        PwmsMotifScanFormatError
            If the file does not strictly follow the MotifScan PWMs format.
        """
        logger.debug(f"Reading MotifScan PWMs from {path}")
        header_pattern = re.compile(r"^>(\S+)\t(\S+)\tPWM$")
        matrix_pattern = re.compile(r"^([ACGT]) \[(.+)\]$")
        cutoff_pattern = re.compile(r"^Cutoff_p(\S+)\t(\S+)")

        pwms = []
        line_num = 0
        # expect_flag: 0=header, 1=matrix, 2=cutoff, 3=cutoff or header
        # 1 header line + 4 matrix line + at least 1 cutoff line
        expect_flag = 0
        with open(path, 'r') as fin:
            for line in fin:
                line_num += 1
                line = line.strip()
                if not line:  # skip blank lines
                    continue

                m_header = header_pattern.match(line)
                m_matrix = matrix_pattern.match(line)
                m_cutoff = cutoff_pattern.match(line)

                # format checker
                if m_header:
                    if expect_flag != 0 and expect_flag != 3:
                        raise PwmsMotifScanFormatError(line_num, line)
                elif m_matrix:
                    if expect_flag != 1:
                        raise PwmsMotifScanFormatError(line_num, line)
                elif m_cutoff:
                    if expect_flag != 2 and expect_flag != 3:
                        raise PwmsMotifScanFormatError(line_num, line)
                else:  # does not match any pattern
                    raise PwmsMotifScanFormatError(line_num, line)

                if m_header:
                    if expect_flag == 3:  # already got a pwm and save it
                        pwm = PositionWeightMatrix(values=values, name=name,
                                                   matrix_id=matrix_id,
                                                   cutoffs=cutoffs)
                        pwms.append(pwm)
                    matrix_id = m_header.group(1)
                    name = m_header.group(2)
                    n_matrix = 0
                    values = []
                    cutoffs = {}
                    expect_flag = 1  # found header, expect matrix line next
                elif m_matrix:
                    base = m_matrix.group(1)
                    if base != bases[n_matrix]:
                        raise PwmsMotifScanFormatError(line_num, line)
                    tmp_values = m_matrix.group(2).split()
                    try:
                        values.append(list(map(float, tmp_values)))
                    except (ValueError, TypeError):
                        raise PwmsMotifScanFormatError(line_num, line)
                    n_matrix += 1
                    if n_matrix == 4:
                        # got 4 matrix line, expect cutoff line next
                        expect_flag = 2
                elif m_cutoff:
                    p = m_cutoff.group(1)
                    cutoff = m_cutoff.group(2)
                    cutoffs[p] = float(cutoff)
                    # got first cutoff, expect cutoff or header next
                    if expect_flag == 2:
                        expect_flag = 3

            # check whether the last matrix is complete
            if expect_flag == 1 or expect_flag == 2:
                raise PwmsMotifScanFormatError(line_num + 1, '')
            if expect_flag == 3:
                pwm = PositionWeightMatrix(values=values, name=name,
                                           matrix_id=matrix_id,
                                           cutoffs=cutoffs)
                pwms.append(pwm)
        self.extend(pwms)
        logger.debug(f"Found {len(pwms)} MotifScan PWMs")

    def set_cutoffs(self, sampling_scores):
        """Set motif score cutoffs given the sampling distributions.

        Parameters
        ----------
        sampling_scores : array_like
            The sampling scores of motifs in a shape of (n_motifs, n_sampling).
        """
        logger.debug(f"Setting motif score cutoffs for PWMs {self.name!r}")
        sampling_scores = np.asarray(sampling_scores)
        if sampling_scores.ndim != 2:
            raise ValueError("sampling scores should have 2 dimensions")
        if len(self) != sampling_scores.shape[0]:
            raise ValueError(
                f"sampling scores are expected to have {len(self)} row(s), "
                f"but got: {sampling_scores.shape[0]} row(s)")
        if sampling_scores.shape[1] < 100:
            raise ValueError(
                "each motif must have at least 100 sampling scores")

        n_bits = min(len(str(sampling_scores.shape[1])), 7)
        for idx, pwm in enumerate(self):
            scores = sampling_scores[idx]
            cutoffs = {}
            for exponent in range(2, n_bits):
                cutoff = np.quantile(scores, 1 - 0.1 ** exponent,
                                     interpolation='lower')
                cutoffs[f'1e-{exponent}'] = cutoff
            pwm.cutoffs = cutoffs


def load_installed_pfms(name):
    """Load a pre-installed motif PFMs set.

    Parameters
    ----------
    name : str
        Name of the pre-installed motif PFMs set to be loaded.

    Return
    ------
    pfms : `MotifPfms`
        Loaded PFMs of the motif set.

    Raises
    ------
    PfmsFileNotFoundError
        If the motif PFMs file does not exists.
    """
    logger.info(f"Loading motif PFMs set {name!r}")
    motif_dir = Config().get_motif_path(name)
    pfms_path = pfms_path_fmt.format(motif_dir, name)
    if os.path.isfile(pfms_path):
        pfms = MotifPfms(name=name)
        pfms.read_pfms(path=pfms_path, format='jaspar')
    else:
        raise PfmsFileNotFoundError(name)
    return pfms


def load_built_pwms(name, genome):
    """Load built motif PWMs.

    Parameters
    ----------
    name : str
        Name of the built motif PWMs set to be loaded.
    genome : str
        Genome assembly name under which these PWMs are built.

    Raises
    ------
    PwmsFileNotFoundError
        If the motif PWMs file does not exists
    """
    logger.info(
        f"Loading motif PWMs set {name!r} under genome {genome!r}")
    motif_dir = Config().get_motif_path(name)
    pwms_path = pwms_path_fmt.format(motif_dir, name, genome)
    pwms = MotifPwms(name=name, genome=genome)
    if os.path.isfile(pwms_path):
        pwms.read_motifscan_pwms(pwms_path)
    else:
        raise PwmsFileNotFoundError(name, genome)
    return pwms



import pytest

from motifscan.motif.matrix import PositionMatrix, PositionFrequencyMatrix, \
    PositionProbabilityMatrix, PositionWeightMatrix


def test_matrix_init():
    matrix = PositionMatrix(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]], name='matrix1',
        matrix_id='matrix_1')
    assert matrix.name == 'matrix1'
    assert matrix.matrix_id == 'matrix_1'
    assert matrix.shape == (4, 3)
    assert matrix.length == 3
    assert len(matrix) == 3
    with pytest.raises(ValueError):
        PositionMatrix([])
    with pytest.raises(ValueError):
        PositionMatrix([1, 2, 3, 4])
    with pytest.raises(ValueError):
        PositionMatrix([[], [], [], []])
    with pytest.raises(ValueError):
        PositionMatrix(['s', 'a', 1, 2])
    with pytest.raises(ValueError):
        PositionMatrix([['s'], ['a'], ['o'], ['e']])


def test_pfm_init():
    with pytest.raises(ValueError):
        PositionFrequencyMatrix([[1], [0.4], [7], [10]])
    with pytest.raises(ValueError):
        PositionFrequencyMatrix([[-1], [4], [7], [10]])
    with pytest.raises(ValueError):
        PositionFrequencyMatrix([[0], [0], [0], [0]])


def test_pfm_to_ppm():
    pfm = PositionFrequencyMatrix([[1, 1], [1, 2], [1, 2], [1, 0]])
    ppm = pfm.to_ppm(normalize=False)
    assert ppm.matrix[0][0] == pytest.approx(0.25)
    assert ppm.matrix[0][1] == pytest.approx(0.2)
    assert ppm.matrix[1][0] == pytest.approx(0.25)
    assert ppm.matrix[1][1] == pytest.approx(0.4)
    assert ppm.matrix[3][1] == pytest.approx(0)
    ppm = pfm.to_ppm(normalize=True, pseudo=0.001)
    assert ppm.matrix[0][0] == pytest.approx(0.25)
    assert ppm.matrix[0][1] == pytest.approx(0.2002)
    assert ppm.matrix[1][0] == pytest.approx(0.25)
    assert ppm.matrix[1][1] == pytest.approx(0.3994)
    assert ppm.matrix[3][1] == pytest.approx(0.001)


def test_ppm_init():
    with pytest.raises(ValueError):
        PositionProbabilityMatrix([[0], [0], [0], [0]])
    with pytest.raises(ValueError):
        PositionProbabilityMatrix([[0], [0.2], [-0.1], [0.9]])
    with pytest.raises(ValueError):
        PositionProbabilityMatrix([[0.3], [0.2], [0.5], [0.3]])


def test_ppm_normalize():
    ppm = PositionProbabilityMatrix(
        [[0.2, 0.2], [0.2, 0.2], [0.3, 0.6], [0.3, 0]])
    with pytest.raises(ValueError):
        ppm.normalize(pseudo=1)
    ppm.normalize(pseudo=0.001)
    assert ppm.matrix[0][0] == pytest.approx(0.2)
    assert ppm.matrix[0][1] == pytest.approx(0.2002)
    assert ppm.matrix[2][0] == pytest.approx(0.3)
    assert ppm.matrix[2][1] == pytest.approx(0.5986)
    assert ppm.matrix[3][1] == pytest.approx(0.001)


def test_ppm_to_pwm():
    ppm = PositionProbabilityMatrix(
        [[0.2, 0.2], [0.2, 0.2], [0.3, 0.6], [0.3, 0]])
    ppm.normalize(pseudo=0.001)
    pwm = ppm.to_pwm()
    assert pwm.matrix[0][0] == pytest.approx(-0.22314)
    assert pwm.matrix[0][1] == pytest.approx(-0.22214)
    assert pwm.matrix[2][0] == pytest.approx(0.18232)
    assert pwm.matrix[2][1] == pytest.approx(0.87313)
    assert pwm.matrix[3][1] == pytest.approx(-5.52146)
    pwm = ppm.to_pwm(bg_freq={'A': 0.22, 'C': 0.23, 'G': 0.28, 'T': 0.27})
    assert pwm.matrix[0][0] == pytest.approx(-0.09531)
    assert pwm.matrix[0][1] == pytest.approx(-0.09431)
    assert pwm.matrix[1][0] == pytest.approx(-0.13976)
    assert pwm.matrix[1][1] == pytest.approx(-0.13876)
    assert pwm.matrix[2][0] == pytest.approx(0.06899)
    assert pwm.matrix[2][1] == pytest.approx(0.75980)
    assert pwm.matrix[3][1] == pytest.approx(-5.59842)


def test_pwm_raw_score():
    pwm = PositionWeightMatrix(
        [[1.35, 0.21, -5.23], [0.07, -0.21, 0.6], [2.15, 2.22, -0.84],
         [-2.64, -1.89, 5.47]])
    assert pwm.max_raw_score == pytest.approx(9.84)
    assert pwm.max_raw_score == pytest.approx(9.84)
    assert pwm.min_raw_score == pytest.approx(-9.76)
    assert pwm.min_raw_score == pytest.approx(-9.76)


def test_pwm_score():
    pwm = PositionWeightMatrix(
        [[1.35, 0.21, -5.23], [0.07, -0.21, 0.6], [2.15, 2.22, -0.84],
         [-2.64, -1.89, 5.47]])
    with pytest.raises(ValueError):
        pwm.score("")
    with pytest.raises(ValueError):
        pwm.score("NNNN")
    assert pwm.score("NNN") == 0
    assert pwm.score("AGT") == pytest.approx(0.918699)
    assert pwm.score("ANT") == pytest.approx(0.693089)
    assert pwm.score("CTA") == pytest.approx(-0.716463)



import logging
import os

import matplotlib as mpl
import numpy as np

mpl.use('Agg')

import matplotlib.pyplot as plt

from motifscan.io.utils import replace_special_char

logger = logging.getLogger(__name__)


def have_same_region_length(regions):
    length = None
    for region in regions:
        if not length:
            length = region.end - region.start
        else:
            if region.end - region.start != length:
                return False
    return True


def have_value_attr(regions):
    for region in regions:
        if region.score is None:
            return False
    return True


def smooth(x, window_len=11):
    if len(x) <= window_len:
        return x
    s = np.r_[x[window_len - 1:0:-1], x, x[-2:-window_len - 1:-1]]
    w = np.hanning(window_len)
    y = np.convolve(w / w.sum(), s, mode='same')
    return y[window_len - 1:-window_len + 1]


def plot_motif_sites_dist(output_dir, regions, pwms, motif_sites,
                          window_size):
    if window_size <= 0:
        if len(regions) == 0:
            logger.error("No regions found for plotting")
            return
        if not have_same_region_length(regions):
            logger.error("Unable to plot when the scanning length is "
                         "different across regions")
            return

    output_dir = os.path.join(output_dir, 'plots')
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    if window_size <= 0:
        window_size = regions[0].end - regions[0].start
    extend = window_size // 2

    for pwm, sites in zip(pwms, motif_sites):
        logger.debug(f"Plotting for {pwm.name}")
        distances = []
        for idx, region in enumerate(regions):
            for site in sites[idx]:
                distances.append(site.start + pwm.length / 2 - region.summit)
        bin_edges = np.arange(-extend - 5, extend + 6, 10)
        freq, _ = np.histogram(distances, bins=bin_edges)
        freq = smooth(freq / len(distances))
        x = []
        for i in range(len(freq)):
            x.append((bin_edges[i] + bin_edges[i + 1]) // 2)
        fig = plt.figure(figsize=(4, 3.5))
        ax = fig.gca()
        ax.bar(x, freq, width=10, color='#4169E1', label=pwm.name)
        ax.legend(loc='upper right', fontsize=8, frameon=False)
        ax.set_xlabel('Distance to Center/Summit', fontsize=8)
        ax.set_ylabel('Fraction', fontsize=8)
        ax.set_xlim(-extend - 5, extend + 5)
        if len(distances) > 0:
            ax.set_ylim(0, 1.2 * max(freq))
        else:
            ax.set_ylim(0, 0.1)
        ax.tick_params(axis='both', which='major', labelsize=8)
        fig.subplots_adjust(left=0.15, right=0.98, bottom=0.15, top=0.95)
        name = replace_special_char(pwm.name)
        path = os.path.join(output_dir, f"{name}_sites_distributions.pdf")
        fig.savefig(path)
        plt.close()


def plot_motif_sites_enrich(output_dir, regions, pwms, motif_sites,
                            motif_sites_control):
    if not have_value_attr(regions):
        logger.error("Unable to plot when some regions have no scores set for "
                     "sorting")
        return
    n_regions_input = len(regions)
    if len(str(n_regions_input)) < 2:
        logger.error(
            f"Too few regions to plot: {n_regions_input}")
        return

    output_dir = os.path.join(output_dir, 'plots')
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    # sort regions by score
    ranked_idx = sorted(range(n_regions_input),
                        key=lambda x: regions[x].score, reverse=True)
    flanking_size = n_regions_input // 100

    for pwm, sites_input, sites_control in zip(pwms, motif_sites,
                                               motif_sites_control):
        logger.debug(f"Plotting for {pwm.name}")
        n_regions_control = len(sites_control)
        n_control = sum([len(sites) > 0 for sites in sites_control])
        ratio_control = n_control / n_regions_control
        if ratio_control == 0:
            ratio_control = 1

        sites_input = np.asarray(sites_input)[ranked_idx]
        fold_changes = []
        has_site_flag = [len(sites) > 0 for sites in sites_input]
        for idx in range(n_regions_input):
            head = max(0, idx - flanking_size)
            tail = min(idx + flanking_size, n_regions_input)
            ratio_input = sum(has_site_flag[head:tail]) / (tail - head)
            fold_changes.append(ratio_input / ratio_control)
        fold_changes = smooth(fold_changes)

        fig = plt.figure(figsize=(4, 3.5))
        ax = fig.gca()
        ax.bar(range(1, n_regions_input + 1), fold_changes, width=1,
               color='#4169E1', label=pwm.name)
        ax.legend(loc='upper right', fontsize=8, frameon=False)
        ax.set_xlabel('Regions Ranked by Score (Descending)', fontsize=8)
        ax.set_ylabel('Fold Change', fontsize=8)
        ax.set_xlim(0, n_regions_input)
        y_max = max(fold_changes)
        if y_max > 0:
            ax.set_ylim(0, 1.2 * y_max)
        else:
            ax.set_ylim(0, 0.1)
        ax.tick_params(axis='both', which='major', labelsize=8)
        fig.subplots_adjust(left=0.15, right=0.98, bottom=0.15, top=0.95)
        name = replace_special_char(pwm.name)
        path = os.path.join(output_dir, f"{name}_sites_enrichment.pdf")
        fig.savefig(path)
        plt.close()



"""
motifscan.exceptions
--------------------

This module contains the exceptions of MotifScan.
"""


class MotifScanError(Exception):
    """Base class for MotifScan errors."""


class InvalidConfigFileError(MotifScanError):
    def __init__(self, path):
        msg = f"Invalid config file: {path}"
        super(MotifScanError, self).__init__(msg)


class RemoteGenomeNotFoundError(MotifScanError):
    def __init__(self, database, assembly):
        msg = f"No genome assembly {assembly!r} in the {database} database"
        super(MotifScanError, self).__init__(msg)


class RemoteGenomeFileNotFoundError(MotifScanError):
    def __init__(self, database, assembly, which):
        msg = f"No {which} file for {assembly!r} in the {database} database"
        super(MotifScanError, self).__init__(msg)


class GenomeNotFoundError(MotifScanError):
    def __init__(self, name):
        msg = f"No such genome assembly: {name!r}"
        super(MotifScanError, self).__init__(msg)


class GenomeFileNotFoundError(MotifScanError):
    def __init__(self, name, which):
        msg = f"No {which} file for assembly {name!r}"
        super(MotifScanError, self).__init__(msg)


class BackgroundFormatError(MotifScanError):
    def __init__(self, line_num, line):
        msg = f"Invalid background format at line {line_num}: {line!r}"
        super(MotifScanError, self).__init__(msg)


class RemoteMotifPFMsNotFoundError(MotifScanError):
    def __init__(self, database, pfms):
        msg = f"No motif PFMs {pfms!r} in the {database} database"
        super(MotifScanError, self).__init__(msg)


class MotifSetNotFoundError(MotifScanError):
    def __init__(self, name):
        msg = f"No such motif set: {name!r}"
        super(MotifScanError, self).__init__(msg)


class PfmsFileNotFoundError(MotifScanError):
    def __init__(self, name):
        msg = f"No PFMs file for motif set {name!r}"
        super(MotifScanError, self).__init__(msg)


class PwmsFileNotFoundError(MotifScanError):
    def __init__(self, name, genome):
        msg = f"No PWMs file for motif set {name!r} under genome {genome}"
        super(MotifScanError, self).__init__(msg)


class PfmsJasparFormatError(MotifScanError):
    def __init__(self, line_num, line):
        msg = f"Invalid JASPAR PFMs format at line {line_num}: {line!r}"
        super(MotifScanError, self).__init__(msg)


class PwmsMotifScanFormatError(MotifScanError):
    def __init__(self, line_num, line):
        msg = f"Invalid MotifScan PWMs format at line {line_num}: {line!r}"
        super(MotifScanError, self).__init__(msg)


class RegionFileFormatError(MotifScanError):
    def __init__(self, format, line_num, line):
        msg = f"Invalid {format} format at line {line_num}: {line!r}"
        super().__init__(msg)



"""
motifscan.motif.databases
-------------------------

Get motifs from public databases.
"""

import logging
import os
import re
import sys
from datetime import datetime

import requests

from motifscan.exceptions import RemoteMotifPFMsNotFoundError

logger = logging.getLogger(__name__)


class JasparDatabase:
    """JASPAR motif database for the JASPAR latest 2020 release, including
    JASPAR CORE (non-redundant/redundant set) and JASPAR collections.

    Attributes
    ----------
    name : str
        Database name, `JASPAR2020` here.
    core_taxons : list of str
        Taxons of JASPAR CORE collection.
    other_collections : list of str
        JASPAR other collections.
    url_core_fmt : str
        URL formatter for motif sets in JASPAR CORE.
    url_other_collections_fmt : str
        URL formatter for motif sets in JASPAR other collections.

    Notes
    -----
        - JASPAR CORE non-redundant set should be used in most situations.
        - JASPAR CORE redundant set includes PFMs released in earlier versions.
    """

    def __init__(self):
        self.name = 'JASPAR2020'
        self.core_taxons = ['vertebrates', 'plants', 'insects', 'nematodes',
                            'fungi', 'urochordates']
        self.other_collections = ['CNE', 'PHYLOFACTS', 'SPLICE', 'POLII',
                                  'FAM', 'PBM', 'PBM_HOMEO', 'PBM_HLH',
                                  'UNVALIDATED']
        url_prefix = "http://jaspar.genereg.net/download/"
        self.url_core_fmt = \
            url_prefix + "CORE/JASPAR2020_CORE_{0}_pfms_jaspar.txt"
        self.url_other_collections_fmt = \
            url_prefix + "collections/JASPAR2020_{0}_pfms_jaspar.txt"
        self._pfms_core = None
        self._pfms_other_collections = self.other_collections

    @property
    def pfms_core(self):
        """Returns available motif PFMs sets in JASPAR CORE.

        Returns
        -------
        list of str
            List of available motif PFMs sets in JASPAR CORE.
        """
        if self._pfms_core is None:
            pfms = []
            for taxon in self.core_taxons:
                pfms.append(f"{taxon}_non-redundant")
                pfms.append(f"{taxon}_redundant")
            self._pfms_core = pfms
        return self._pfms_core

    @property
    def pfms_other_collections(self):
        """Returns available motif PFMs sets in JASPAR other collections.

        Returns
        -------
        list of str
            List of available motif PFMs sets in JASPAR other collections.
        """
        return self._pfms_other_collections

    @staticmethod
    def _download_pfms(pfms_url, download_dir):
        """Download PFMs from `pfms_url` and save to `download_dir`."""
        if not os.path.isdir(os.path.dirname(download_dir)):
            os.makedirs(download_dir)
        base_name = os.path.basename(pfms_url)
        dst = os.path.join(download_dir, base_name)
        try:
            logger.debug(f"Downloading {pfms_url}")
            r = requests.get(pfms_url, stream=True)
            r.raise_for_status()
            with open(dst, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024):
                    f.write(chunk)
        except requests.HTTPError as e:
            logger.error(f"Failed to download due to an HTTPError: {e}")
            sys.exit(1)
        return dst

    @staticmethod
    def _write_readme(database, pfms_name, download_dir):
        logger.debug("Writing the README file")
        readme_file = os.path.join(download_dir, "README")
        time_now = datetime.now().strftime("%Y-%m-%d %H:%M")
        with open(readme_file, 'w') as f_out:
            f_out.write(f"{database}\t{pfms_name}\tDownloaded at {time_now}\n")

    def download_core(self, pfms_name, download_dir):
        """Download a motif PFMs set from JASPAR CORE.

        Parameters
        ----------
        pfms_name : str
            Motif PFMs set name.
        download_dir : str
            Directory to write downloaded file.

        Returns
        -------
        dst_pfms : str
            Path of saved PFMs file

        Raises
        ------
        RemotePfmsNotFoundError
            If the specified PFMs is not in JASPAR CORE.
        """
        m = re.match(r"^([a-z]+)_(non-)?redundant$", pfms_name)
        if m and m.group(1) in self.core_taxons:
            logger.info(
                f"Downloading motif PFMs set {pfms_name!r} from JASPAR CORE")
            pfms_url = self.url_core_fmt.format(pfms_name)
            dst_pfms = self._download_pfms(pfms_url, download_dir)
            self._write_readme("JASPAR2020_CORE", pfms_name, download_dir)
        else:
            raise RemoteMotifPFMsNotFoundError("JASPAR CORE", pfms_name)
        return dst_pfms

    def download_other_collections(self, pfms_name, download_dir):
        """Download motif PFMs from JASPAR collections (other than CORE).

        Parameters
        ----------
        pfms_name : str
            Motif PFMs name.
        download_dir : str
            Directory to write downloaded file.

        Returns
        -------
        dst_pfms : str
            Path of saved PFMs file

        Raises
        ------
        RemotePfmsNotFoundError
            If the specified PFMs is not in JASPAR collections.
        """
        m = re.match(r"^([A-Z_]+)", pfms_name)
        if m and m.group(1) in self.other_collections:
            logger.info(f"Downloading motif PFMs set {pfms_name!r} from "
                        f"JASPAR Collections")
            pfms_url = self.url_other_collections_fmt.format(pfms_name)
            dst_pfms = self._download_pfms(pfms_url, download_dir)
            self._write_readme("JASPAR2020_Collections", pfms_name,
                               download_dir)
        else:
            raise RemoteMotifPFMsNotFoundError("JASPAR Collections", pfms_name)
        return dst_pfms

    @staticmethod
    def get_motif_info(matrix_id):
        """Get detailed information about a JASPAR motif.

        Parameters
        ----------
        matrix_id : str
            The matrix id of the JASPAR motif.

        Returns
        -------
        motif_info : dict
            A dict object contains the motif information.
        """
        motif_info = {}
        try:
            logger.debug(f"Getting motif info: {matrix_id}")
            url = f"http://jaspar.genereg.net/api/v1/matrix/{matrix_id}/"
            r = requests.get(url)
            r.raise_for_status()
            motif_info = r.json()
        except requests.HTTPError as e:
            logger.error(f"Failed to get motif info from JASPAR: {e}")
        return motif_info



import os
import shutil

import pytest

from motifscan.config import Config


@pytest.fixture(scope='session')
def data_dir():
    """Return the directory of data files."""
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


@pytest.fixture(scope='session')
def tmp_dir(data_dir):
    """Generate a temporary output directory to write output files."""
    tmp_dir = os.path.join(data_dir, 'motifscan_tmp_output')
    if os.path.isdir(tmp_dir):
        shutil.rmtree(tmp_dir)
    os.makedirs(tmp_dir)
    yield tmp_dir
    if os.path.isdir(tmp_dir):
        shutil.rmtree(tmp_dir)


@pytest.fixture(scope='function')
def config(tmp_dir):
    """Returns a config instance to test the configuration of MotifScan."""
    return Config(os.path.join(tmp_dir, '.motifscanrc'))


@pytest.fixture(scope='session')
def genome_root(data_dir):
    """Returns the path of test genome."""
    return os.path.join(data_dir, 'genomes')


@pytest.fixture(scope='session')
def motif_root(data_dir):
    """Returns the path of test motif set."""
    return os.path.join(data_dir, 'motifs')


@pytest.fixture(scope='session')
def region_root(data_dir):
    """Returns the path of test regions."""
    return os.path.join(data_dir, 'regions')



"""
motifscan.cli.motif
-------------------

Command line interface for the 'motif' subcommand.
"""

import logging
import os
import shutil
import sys

from motifscan.config import Config
from motifscan.exceptions import RemoteMotifPFMsNotFoundError, \
    MotifSetNotFoundError
from motifscan.genome import Genome
from motifscan.io.utils import merge_files
from motifscan.motif import pfms_path_fmt, MotifPwms, load_installed_pfms
from motifscan.motif.databases import JasparDatabase
from motifscan.motif.score import motif_score

logger = logging.getLogger(__name__)


def run(args, config_file=None):
    if args.list:
        config = Config(config_file)
        for name, _ in config.list_motif_sets():
            print(name)
        return
    if args.list_remote:
        database = JasparDatabase()
        if args.database == 'jaspar_core':
            remote_sets = database.pfms_core
            for name in remote_sets:
                print(f"{name:25}\t{database.name + '_CORE'}")
        else:
            remote_sets = database.pfms_other_collections
            for name in remote_sets:
                print(f"{name:20}\t{database.name + '_Collections'}")
        return
    if args.install:
        install_motif(args, config_file)
        return
    if args.build:
        build_motif(args, config_file)
        return
    if args.uninstall:
        uninstall_motif(args, config_file)
        return


def install_motif(args, config_file=None):
    config = Config(config_file)
    if config.has_motif_set(args.name):
        logger.error(f"Motif set {args.name!r} already exists!")
        sys.exit(1)

    motif_dir = os.path.abspath(
        args.output_dir or os.path.join(config.get_motif_dir(), args.name))
    logger.info(f"Installing motif set {args.name!r} into {motif_dir}")
    if not os.path.isdir(motif_dir):
        os.makedirs(motif_dir)
    if os.listdir(motif_dir):
        logger.error("Directory not empty! Please specify another directory "
                     "or delete files under it.")
        sys.exit(1)

    pfms_path = pfms_path_fmt.format(motif_dir, args.name)

    if args.remote:
        try:
            db = JasparDatabase()
            if args.database == 'jaspar_core':
                dst_pfms = db.download_core(args.remote, motif_dir)
            else:
                dst_pfms = db.download_other_collections(args.remote,
                                                         motif_dir)
            logger.debug(
                f"Renaming downloaded file to {os.path.basename(pfms_path)}")
            shutil.move(dst_pfms, pfms_path)
        except RemoteMotifPFMsNotFoundError as e:
            logger.error(e)
            sys.exit(1)
    else:
        logger.info("Copying the PFMs file(s)")
        merge_files(args.pfm_files, pfms_path)

    logger.info("Updating the config file")
    config.set_motif_path(args.name, motif_dir)
    config.write()
    logger.info("Successfully installed!")
    if args.genome:
        build_motif(args, config_file)


def build_motif(args, config_file=None):
    if args.build:
        name = args.build
    else:
        name = args.name
    logger.info(
        f"Building motif set {name!r} for genome assembly {args.genome!r}")
    genome = Genome(args.genome)
    pfms = load_installed_pfms(name)

    logger.info("Converting motif PFMs to PWMs")
    max_length = 0
    pwms = MotifPwms(name=name, genome=genome.name)
    for pfm in pfms:
        max_length = max(max_length, pfm.length)
        pwm = pfm.to_ppm().to_pwm(genome.bg_freq)
        pwms.append(pwm)

    logger.info("Random sampling sequences")
    random_sequences = list(genome.random_sequences(args.n_random, max_length,
                                                    args.seed))
    logger.info("Scanning motifs on the sampled sequences")
    sampling_scores = []
    for idx, pwm in enumerate(pwms):
        logger.info(f"Scanning {pwm.name} [{idx + 1}/{len(pwms)}]")
        matrix = pwm.matrix.tolist()
        scores = motif_score(matrix, pwm.length, pwm.max_raw_score,
                             random_sequences)
        sampling_scores.append(scores)
    logger.info("Setting motif score cutoffs")
    pwms.set_cutoffs(sampling_scores)
    pwms.save_built_pwms()
    logger.info("Successfully built!")


def uninstall_motif(args, config_file=None):
    try:
        config = Config(config_file)
        path = config.get_motif_path(args.uninstall)
    except MotifSetNotFoundError as e:
        logger.error(e)
        sys.exit(1)

    logger.info(f"Uninstalling motif set {args.uninstall!r}")
    if os.path.isdir(path):
        logger.info(f"Removing files under {path}")
        try:
            shutil.rmtree(path)
        except Exception as e:
            logger.error(f"Failed to remove the motif directory: {e}")
            sys.exit(1)

    logger.info("Updating the config file")
    config.remove_motif_path(args.uninstall)
    config.write()
    logger.info("Successfully uninstalled!")



import os

import pytest

from motifscan.exceptions import RegionFileFormatError
from motifscan.region import load_motifscan_regions


def test_unsupported_format(region_root):
    with pytest.raises(ValueError):
        load_motifscan_regions(os.path.join(region_root, 'test_regions.bed'),
                               format='unknown_format')


def test_invalid_format(region_root):
    with pytest.raises(RegionFileFormatError):
        load_motifscan_regions(os.path.join(region_root, 'test_regions.bed'),
                               format='bed3-summit')


def test_bed_parser(region_root):
    regions = load_motifscan_regions(
        os.path.join(region_root, 'test_regions.bed'),
        format='bed')
    assert len(regions) == 4
    assert regions[0].score == 100
    assert regions[-1].score == None


def test_bed3_summit_parser(region_root):
    regions = load_motifscan_regions(
        os.path.join(region_root, 'test_regions_summit.bed'),
        format='bed3-summit')
    assert len(regions) == 4


def test_macs_parser(region_root):
    regions = load_motifscan_regions(
        os.path.join(region_root, 'test_regions_macs.xls'),
        format='macs')
    assert len(regions) == 9
    assert regions[0].score == 100.54


def test_macs2_parser(region_root):
    regions = load_motifscan_regions(
        os.path.join(region_root, 'test_regions_macs2.xls'),
        format='macs2')
    assert len(regions) == 10
    assert regions[0].score == 33.13990


def test_narrowpeak_parser(region_root):
    regions = load_motifscan_regions(
        os.path.join(region_root, 'test_regions.narrowPeak'),
        format='narrowpeak')
    assert len(regions) == 10
    assert regions[0].score == 331


def test_broadpeak_parser(region_root):
    regions = load_motifscan_regions(
        os.path.join(region_root, 'test_regions.broadPeak'),
        format='broadpeak')
    assert len(regions) == 10
    assert regions[0].score == 331


def test_manorm_parser(region_root):
    regions = load_motifscan_regions(
        os.path.join(region_root, 'test_regions_manorm.xls'),
        format='manorm')
    assert len(regions) == 4
    assert regions[0].score == 6.77003



import os

import pytest

from motifscan.config import Config, user_rc_path, user_genome_dir, \
    user_motif_dir
from motifscan.exceptions import InvalidConfigFileError, GenomeNotFoundError, \
    MotifSetNotFoundError


def test_invalid_config(data_dir):
    with pytest.raises(InvalidConfigFileError):
        Config(os.path.join(data_dir, 'invalid.motifscanrc'))


def test_config_init(config):
    assert len(config._config.sections()) == 3
    assert config._config.has_section('motifscan')
    assert config._config.has_section('genome')
    assert config._config.has_section('motif')
    assert config._config.get('motifscan', 'genome_dir') == user_genome_dir
    assert config._config.get('motifscan', 'motif_dir') == user_motif_dir
    config = Config(path=None)
    assert config.path == user_rc_path


def test_get_genome_dir(config):
    assert config.get_genome_dir() == user_genome_dir


def test_set_genome_dir(config):
    path = 'test_dir'
    config.set_genome_dir(path)
    assert config.get_genome_dir() == path


def test_get_motif_dir(config):
    assert config.get_motif_dir() == user_motif_dir


def test_set_motif_dir(config):
    path = 'test_dir'
    config.set_motif_dir(path)
    assert config.get_motif_dir() == path


def test_set_genome_path(config):
    name = 'hg19'
    path = 'test_dir'
    config.set_genome_path(name, path)
    assert config._config.get('genome', name) == path


def test_get_genome_path(config):
    name = 'hg19'
    path = 'test_dir'
    config.set_genome_path(name, path)
    assert config.get_genome_path(name) == path
    with pytest.raises(GenomeNotFoundError):
        assert config.get_genome_path('mm9')


def test_remove_genome_path(config):
    name = 'hg19'
    path = 'test_dir'
    config.set_genome_path(name, path)
    assert config.remove_genome_path(name)
    with pytest.raises(GenomeNotFoundError):
        assert config.remove_genome_path(name)


def test_list_genome_assemblies(config):
    config.set_genome_path('hg19', 'test_dir1')
    config.set_genome_path('hg18', 'test_dir2')
    assert list(config.list_genome_assemblies()) == [('hg19', 'test_dir1'),
                                                     ('hg18', 'test_dir2')]


def test_has_genome_assembly(config):
    name = 'hg19'
    path = 'test_dir'
    config.set_genome_path(name, path)
    assert config.has_genome_assembly(name)
    assert not config.has_genome_assembly('mm9')


def test_set_motif_path(config):
    name = 'motif_set'
    path = 'test_dir'
    config.set_motif_path(name, path)
    assert config._config.get('motif', name) == path


def test_get_motif_path(config):
    name = 'motif_set'
    path = 'test_dir'
    config.set_motif_path(name, path)
    assert config.get_motif_path(name) == path
    with pytest.raises(MotifSetNotFoundError):
        assert config.get_motif_path('motif_set1')


def test_remove_motif_path(config):
    name = 'motif_set'
    path = 'test_dir'
    config.set_motif_path(name, path)
    assert config.remove_motif_path(name)
    with pytest.raises(MotifSetNotFoundError):
        assert config.remove_motif_path(name)


def test_list_motif_sets(config):
    config.set_motif_path('motif_set1', 'test_dir1')
    config.set_motif_path('motif_set2', 'test_dir2')
    assert list(config.list_motif_sets()) == [('motif_set1', 'test_dir1'),
                                              ('motif_set2', 'test_dir2')]


def test_has_motif_set(config):
    name = 'motif_set'
    path = 'test_dir'
    config.set_motif_path(name, path)
    assert config.has_motif_set(name)
    assert not config.has_motif_set('motif_set1')


def test_write(config):
    config.set_genome_dir("genome_root_dir")
    config.set_motif_dir("motif_root_dir")
    config.set_genome_path('hg19', 'test_dir1')
    config.set_genome_path('mm9', 'test_dir2')
    config.set_motif_path('motif_set1', 'test_dir3')
    config.set_motif_path('motif_set2', 'test_dir4')
    config.write()
    assert os.path.isfile(config.path)
    fin = open(config.path, 'r')
    assert fin.read() == "[motifscan]\n" \
                         "genome_dir = genome_root_dir\n" \
                         "motif_dir = motif_root_dir\n\n" \
                         "[genome]\n" \
                         "hg19 = test_dir1\n" \
                         "mm9 = test_dir2\n\n" \
                         "[motif]\n" \
                         "motif_set1 = test_dir3\n" \
                         "motif_set2 = test_dir4\n\n"



"""
motifscan.io
------------
I/O functions for MotifScan.
"""

import os

from motifscan.io.utils import replace_special_char


def write_sites_table(output_dir, pwms, regions, motif_sites):
    """Write the motif sites number/score summary table."""
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    path_num = os.path.join(output_dir, 'motif_sites_number.xls')
    path_score = os.path.join(output_dir, 'motif_sites_score.xls')
    with open(path_num, 'w') as f_num, open(path_score, 'w') as f_score:
        name_fields = '\t'.join([pwm.name for pwm in pwms])
        f_num.write(f"chr\tstart\tend\t{name_fields}\n")
        f_score.write(f"chr\tstart\tend\t{name_fields}\n")
        for idx, region in enumerate(regions):
            n_sites = []
            scores = []
            for sites in motif_sites:  # sites_by_motif
                num = len(sites[idx])  # sites_by_region
                n_sites.append(num)
                if num == 0:
                    scores.append('NA')
                else:
                    scores.append(max([site.score for site in sites[idx]]))
            num_fields = '\t'.join(map(str, n_sites))
            score_fields = '\t'.join(map(str, scores))
            f_num.write(f"{region.chrom}\t{region.start + 1}\t{region.end}\t"
                        f"{num_fields}\n")
            f_score.write(f"{region.chrom}\t{region.start + 1}\t{region.end}\t"
                          f"{score_fields}\n")


def write_sites_bed(output_dir, pwms, regions, motif_sites):
    """Write motif sites bed file(s)."""
    output_dir = os.path.join(output_dir, 'motif_sites')
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    for pwm, sites in zip(pwms, motif_sites):
        name = replace_special_char(pwm.name)
        path = os.path.join(output_dir, f"{name}_sites.bed")
        with open(path, 'w') as f_out:
            for idx, region in enumerate(regions):
                for site in sites[idx]:
                    f_out.write(f"{region.chrom}\t{site.start}\t"
                                f"{site.start + pwm.length}\t.\t"
                                f"{site.score}\t{site.strand}\n")


def write_enrich_table(output_dir, enrichment_results):
    """Write the motif enrichment results table."""
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    path = os.path.join(output_dir, 'motif_enrichment.xls')
    # sort by enriched p value and then fold change if p value ties
    enrichment_results.sort(key=lambda x: (x.p_enriched, -x.fold_change))
    with open(path, 'w') as f_out:
        f_out.write("Motif\tNum_input_regions\tNum_control_regions\t"
                    "Fold_change\tEnriched_P_value\tDepleted_P_value\t"
                    "Corrected_P_value\n")
        for res in enrichment_results:
            f_out.write(f"{res.name}\t{res.n_input}\t{res.n_control}\t"
                        f"{res.fold_change}\t{res.p_enriched}\t"
                        f"{res.p_depleted}\t{res.p_corrected}\n")



import os

import pytest

from motifscan.exceptions import MotifSetNotFoundError, \
    PfmsJasparFormatError, PwmsMotifScanFormatError
from motifscan.motif import MotifPfms, MotifPwms, load_installed_pfms, \
    load_built_pwms
from motifscan.motif.matrix import PositionFrequencyMatrix, \
    PositionWeightMatrix


def test_pfms_init(motif_root):
    with pytest.raises(ValueError):
        MotifPfms(
            pfms=[PositionFrequencyMatrix([[1], [2], [3], [4]]), 's', True])
    name = 'motif_set'
    pfm = PositionFrequencyMatrix([[1], [2], [3], [4]])
    pfms = MotifPfms(pfms=[pfm], name=name)
    assert pfms.name == name
    assert list(pfms) == [pfm]
    assert len(pfms) == 1
    with pytest.raises(MotifSetNotFoundError):
        load_installed_pfms('uninstalled_motif_set')


def test_read_jaspar_pfms(motif_root):
    pfms = MotifPfms()
    pfms.read_pfms(os.path.join(motif_root, 'test', 'test_pfms.jaspar'),
                   format='jaspar')
    assert len(pfms) == 2
    with pytest.raises(ValueError):
        pfms.read_pfms(os.path.join(motif_root, 'test', 'test_pfms.jaspar'),
                       format='jas')


def test_pfms_jaspar_format(motif_root):
    pfms = MotifPfms()
    with pytest.raises(PfmsJasparFormatError):
        pfms.read_pfms(
            os.path.join(motif_root, 'bad', 'test_pfms_bad1.jaspar'),
            format='jaspar')
    with pytest.raises(PfmsJasparFormatError):
        pfms.read_pfms(
            os.path.join(motif_root, 'bad', 'test_pfms_bad2.jaspar'),
            format='jaspar')
    with pytest.raises(PfmsJasparFormatError):
        pfms.read_pfms(
            os.path.join(motif_root, 'bad', 'test_pfms_bad3.jaspar'),
            format='jaspar')
    with pytest.raises(PfmsJasparFormatError):
        pfms.read_pfms(
            os.path.join(motif_root, 'bad', 'test_pfms_bad4.jaspar'),
            format='jaspar')


def test_pwms_init(motif_root):
    with pytest.raises(ValueError):
        MotifPwms(
            pwms=[PositionWeightMatrix([[1], [2], [3], [4]]), 's', True])
    pwm = PositionWeightMatrix([[1], [2], [3], [4]])
    name = 'motif_set'
    pwms = MotifPwms(pwms=[pwm], name=name)
    assert len(pwms) == 1
    assert pwms.name == name
    pwms = MotifPwms(name=name)
    pwms.read_motifscan_pwms(
        os.path.join(motif_root, 'test', 'test_pwms.motifscan'))
    assert len(pwms) == 2
    with pytest.raises(MotifSetNotFoundError):
        load_built_pwms('uninstalled_motif_set', 'hg19')


def test_pwms_motifscan_format(motif_root):
    pwms = MotifPwms()
    with pytest.raises(PwmsMotifScanFormatError):
        pwms.read_motifscan_pwms(
            os.path.join(motif_root, 'bad', 'test_pwms_bad1.motifscan'))
    with pytest.raises(PwmsMotifScanFormatError):
        pwms.read_motifscan_pwms(
            os.path.join(motif_root, 'bad', 'test_pwms_bad2.motifscan'))
    with pytest.raises(PwmsMotifScanFormatError):
        pwms.read_motifscan_pwms(
            os.path.join(motif_root, 'bad', 'test_pwms_bad3.motifscan'))
    with pytest.raises(PwmsMotifScanFormatError):
        pwms.read_motifscan_pwms(
            os.path.join(motif_root, 'bad', 'test_pwms_bad4.motifscan'))
    with pytest.raises(PwmsMotifScanFormatError):
        pwms.read_motifscan_pwms(
            os.path.join(motif_root, 'bad', 'test_pwms_bad5.motifscan'))
    with pytest.raises(PwmsMotifScanFormatError):
        pwms.read_motifscan_pwms(
            os.path.join(motif_root, 'bad', 'test_pwms_bad6.motifscan'))


def test_pwms_write_motifscan_pwms(tmp_dir):
    pwm = PositionWeightMatrix([[1], [2], [3], [4]], name='motif1',
                               matrix_id='id1',
                               cutoffs={'1e-2': 0.1, '1e-3': 0.4})
    name = 'motif_set'
    pwms = MotifPwms(pwms=[pwm], name=name)
    pwms_path = os.path.join(tmp_dir, 'test_pwm.motifscan')
    pwms.write_motifscan_pwms(pwms_path)
    assert os.path.isfile(pwms_path)


def test_set_cutoffs():
    pwm = PositionWeightMatrix([[1], [2], [3], [4]])
    name = 'motif_set'
    pwms = MotifPwms(pwms=[pwm], name=name)
    pwms.set_cutoffs([range(1, 1000001)])
    assert len(pwm.cutoffs) == 5
    assert pwm.cutoffs['1e-2'] == 990000
    assert pwm.cutoffs['1e-3'] == 999000
    assert pwm.cutoffs['1e-4'] == 999900
    assert pwm.cutoffs['1e-5'] == 999990
    assert pwm.cutoffs['1e-6'] == 999999
    with pytest.raises(ValueError):
        pwms.set_cutoffs([[1], [2]])
    with pytest.raises(ValueError):
        pwms.set_cutoffs([1, 2, 3])
    with pytest.raises(ValueError):
        pwms.set_cutoffs([[1, 2, 3]])



import os

import pytest

from motifscan.genome import Genome, cal_bg_freq, read_bg_freq, write_bg_freq
from motifscan.exceptions import GenomeFileNotFoundError, \
    BackgroundFormatError


def test_genome_init(genome_root):
    genome_path = os.path.join(genome_root, 'test')
    genome = Genome(name='test', path=genome_path)
    assert genome.name == 'test'
    assert genome.path == genome_path
    assert genome.chroms == ['chr1', 'chr2', 'chrM', 'chrX']
    assert genome.chrom_sizes['chr1'] == 10
    assert genome.chrom_sizes['chr2'] == 17
    assert genome.chrom_sizes['chrM'] == 15
    assert genome.chrom_sizes['chrX'] == 16
    assert genome.fetch_sequence('chr1', 0, 10) == 'AaTtCcGgNn'
    assert genome.fetch_sequence('chr2', 0, 17) == 'AAAaCCccTTtGNNNNN'
    assert genome.fetch_sequence('chrM', 0, 15) == 'AaaaaAAAAAAAnnn'
    assert genome.fetch_sequence('chrX', 0, 16) == 'AAACCTACNNTnggAC'
    seq_random = list(genome.random_sequences(n_times=3, length=5))
    assert len(seq_random) == 3
    seq_random = list(
        genome.random_sequences(n_times=3, length=5, random_seed=1))
    assert seq_random == ['AAAAA', 'AaTtC', 'AAAaC']
    genome.close()
    assert genome.fa.closed
    with pytest.raises(GenomeFileNotFoundError):
        Genome(name='unknown_genome', path=genome_path)


def test_genome_cal_bg_freq(genome_root):
    fasta_path = os.path.join(genome_root, 'test', 'test.fa')
    bg_freq = cal_bg_freq(fasta_path, skip_non_autosomes=True)
    assert sorted(bg_freq.keys()) == ['A', 'C', 'G', 'T']
    assert bg_freq['A'] == pytest.approx(0.3)
    assert bg_freq['C'] == pytest.approx(0.3)
    assert bg_freq['G'] == pytest.approx(0.15)
    assert bg_freq['T'] == pytest.approx(0.25)
    bg_freq = cal_bg_freq(fasta_path, skip_non_autosomes=False)
    assert bg_freq['A'] == pytest.approx(0.51111)
    assert bg_freq['C'] == pytest.approx(0.22222)
    assert bg_freq['G'] == pytest.approx(0.11111)
    assert bg_freq['T'] == pytest.approx(0.15556)


def test_read_bg_freq(genome_root):
    bg_freq_path = os.path.join(genome_root, 'test', 'test_bg_freq.txt')
    bg_freq = read_bg_freq(bg_freq_path)
    assert sorted(bg_freq.keys()) == ['A', 'C', 'G', 'T']
    assert bg_freq['A'] == 0.3
    assert bg_freq['C'] == 0.3
    assert bg_freq['G'] == 0.15
    assert bg_freq['T'] == 0.25


def test_bg_freq_format(genome_root):
    genome_path = os.path.join(genome_root, 'bad')
    with pytest.raises(BackgroundFormatError):
        read_bg_freq(os.path.join(genome_path, 'test_bg_freq_bad1.txt'))
    with pytest.raises(BackgroundFormatError):
        read_bg_freq(os.path.join(genome_path, 'test_bg_freq_bad2.txt'))


def test_write_bg_freq(genome_root, tmp_dir):
    fasta_path = os.path.join(genome_root, 'test', 'test.fa')
    bg_freq = cal_bg_freq(fasta_path, skip_non_autosomes=True)
    bg_freq_path = os.path.join(tmp_dir, 'test_bg_freq.txt')
    write_bg_freq(bg_freq_path, bg_freq)
    bg_freq = read_bg_freq(bg_freq_path)
    assert bg_freq['A'] == 0.3
    assert bg_freq['C'] == 0.3
    assert bg_freq['G'] == 0.15
    assert bg_freq['T'] == 0.25



"""
motifscan.cli.scan
------------------

Command line interface for the 'scan' subcommand.
"""

import logging
import sys

from motifscan import __version__
from motifscan.genome import Genome
from motifscan.io import write_sites_table, write_sites_bed, write_enrich_table
from motifscan.motif import load_built_pwms
from motifscan.plot import plot_motif_sites_dist, plot_motif_sites_enrich
from motifscan.region import load_motifscan_regions
from motifscan.region.utils import subset_by_location, generate_control_regions
from motifscan.scanner import Scanner
from motifscan.stats import motif_enrichment

logger = logging.getLogger(__name__)


def run(args):
    logger.info(f"Running MotifScan {__version__}")
    logger.info("===== Loading data =====")
    genome = Genome(name=args.genome)
    pwms = load_built_pwms(name=args.motif, genome=args.genome)
    regions = load_motifscan_regions(path=args.input_file,
                                     format=args.input_format)
    if args.location is not None:
        logger.info(f"Extracting input regions located at {args.location}")
        if genome.genes is None:
            logger.error(f"Unable to extract without gene annotations.")
            sys.exit(1)
        regions = subset_by_location(
            regions=regions, genes=genome.genes, location=args.location,
            upstream=args.upstream, downstream=args.downstream)
        logger.info(f"Extracted {len(regions)} {args.location} regions")

    logger.info("===== Scanning motifs =====")
    logger.info("Fetching the sequences of input regions")
    scanner = Scanner(
        genome=genome, regions=regions, window_size=args.window_size,
        strand=args.strand, p_value=args.p_value, remove_dup=True,
        n_threads=args.n_threads)
    logger.info("Scanning motifs...")
    motif_sites = scanner.scan_motifs(pwms=pwms)

    logger.info("Saving the result tables")
    write_sites_table(output_dir=args.output_dir, pwms=pwms, regions=regions,
                      motif_sites=motif_sites)
    if args.report_site:
        logger.info("Saving the coordinates of detected motif sites")
        write_sites_bed(output_dir=args.output_dir, pwms=pwms, regions=regions,
                        motif_sites=motif_sites)

    if not args.no_enrich:
        logger.info("===== Motif Enrichment =====")
        if args.control_file:
            logger.info("Loading user specified control regions")
            control_regions = load_motifscan_regions(
                path=args.control_file, format=args.control_format)
            if args.location is not None:
                logger.info(
                    f"Extracting control regions located at {args.location}")
                control_regions = subset_by_location(
                    regions=control_regions, genes=genome.genes,
                    location=args.location,
                    upstream=args.upstream, downstream=args.downstream)
                logger.info(
                    f"Extracted {len(control_regions)} {args.location} "
                    f"control regions")
        else:
            logger.info(f"Generating random control regions")
            control_regions = generate_control_regions(
                n_random=args.n_random, regions=regions,
                chrom_size=genome.chrom_sizes, genes=genome.genes,
                random_seed=args.seed)
        logger.info("Fetching the sequences of control regions")
        scanner_control = Scanner(
            genome=genome, regions=control_regions,
            window_size=args.window_size, strand=args.strand,
            p_value=args.p_value, remove_dup=True, n_threads=args.n_threads)
        logger.info("Scanning motifs...")
        motif_sites_control = scanner_control.scan_motifs(pwms=pwms)

        logger.info("Performing motif enrichment analysis")
        enrichment_results = motif_enrichment(
            pwms=pwms, motif_sites=motif_sites,
            motif_sites_control=motif_sites_control)
        logger.info("Saving the motif enrichment table")
        write_enrich_table(output_dir=args.output_dir,
                           enrichment_results=enrichment_results)

    if args.plot_dist:
        logger.info("Plotting the distributions of detected motif sites")
        plot_motif_sites_dist(
            output_dir=args.output_dir, regions=regions, pwms=pwms,
            motif_sites=motif_sites, window_size=args.window_size)
        if not args.no_enrich:
            logger.info("Plotting the enrichment of detected motif sites")
            plot_motif_sites_enrich(
                output_dir=args.output_dir, regions=regions, pwms=pwms,
                motif_sites=motif_sites,
                motif_sites_control=motif_sites_control)

    logger.info("===== MotifScan Finished =====")



import os

import pytest

from motifscan.cli.config import run
from motifscan.cli.main import configure_parser_main
from motifscan.config import Config
from motifscan.exceptions import GenomeNotFoundError, MotifSetNotFoundError

parser = configure_parser_main()


def test_cli_config_show(tmp_dir, capsys):
    config_file = os.path.join(tmp_dir, "test_cli_config.motifscanrc")
    config = Config(config_file)
    config.set_genome_dir("/path/to/genome/root")
    config.set_motif_dir("/path/to/motif/root")
    config.set_genome_path("hg19", "/path/to/genome")
    config.set_motif_path("motif_set", "/path/to/motif")
    config.write()

    args = parser.parse_args(["config", "--show"])
    run(args=args, config_file=config_file)
    captured = capsys.readouterr()
    assert captured.out == "[motifscan]\n" \
                           "genome_dir: /path/to/genome/root\n" \
                           "motif_dir: /path/to/motif/root\n\n" \
                           "[genome]\n" \
                           "hg19: /path/to/genome\n\n" \
                           "[motif]\n" \
                           "motif_set: /path/to/motif\n"


def test_cli_config_set_default_genome(tmp_dir):
    args = parser.parse_args(
        ["config", "--set-default-genome", "/path/to/genome/root"])
    config_file = os.path.join(tmp_dir, "test_cli_config.motifscanrc")
    run(args=args, config_file=config_file)

    config = Config(config_file)
    assert config.get_genome_dir() == "/path/to/genome/root"


def test_cli_config_set_default_motif(tmp_dir):
    args = parser.parse_args(
        ["config", "--set-default-motif", "/path/to/motif/root"])
    config_file = os.path.join(tmp_dir, "test_cli_config.motifscanrc")
    run(args=args, config_file=config_file)

    config = Config(config_file)
    assert config.get_motif_dir() == "/path/to/motif/root"


def test_cli_config_set_genome(tmp_dir):
    args = parser.parse_args(
        ["config", "--set-genome", "hg19", "/path/to/genome"])
    config_file = os.path.join(tmp_dir, 'test_cli_config.motifscanrc')
    run(args=args, config_file=config_file)

    config = Config(config_file)
    assert config.has_genome_assembly("hg19")
    assert config.get_genome_path("hg19") == "/path/to/genome"


def test_cli_config_get_genome(tmp_dir, capsys):
    config_file = os.path.join(tmp_dir, "test_cli_config.motifscanrc")
    config = Config(config_file)
    config.set_genome_path("hg19", "/path/to/genome")
    config.write()

    args = parser.parse_args(["config", "--get-genome", "hg19"])
    run(args=args, config_file=config_file)
    captured = capsys.readouterr()
    assert captured.out == "/path/to/genome\n"

    with pytest.raises(SystemExit):
        args = parser.parse_args(["config", "--get-genome", "hg38"])
        run(args=args, config_file=config_file)


def test_cli_config_rm_genome(tmp_dir):
    config_file = os.path.join(tmp_dir, "test_cli_config.motifscanrc")
    config = Config(config_file)
    config.set_genome_path("hg19", "/path/to/genome")
    config.write()

    args = parser.parse_args(["config", "--rm-genome", "hg19"])
    run(args=args, config_file=config_file)
    config = Config(config_file)
    assert not config.has_genome_assembly("hg19")

    with pytest.raises(GenomeNotFoundError):
        config.get_genome_path("hg19")
    with pytest.raises(SystemExit):
        run(args=args, config_file=config_file)


def test_cli_config_set_motif(tmp_dir):
    args = parser.parse_args(
        ["config", "--set-motif", "motif_set", "/path/to/motif"])
    config_file = os.path.join(tmp_dir, 'test_cli_config.motifscanrc')
    run(args=args, config_file=config_file)

    config = Config(config_file)
    assert config.has_motif_set("motif_set")
    assert config.get_motif_path("motif_set") == "/path/to/motif"


def test_cli_config_get_motif(tmp_dir, capsys):
    config_file = os.path.join(tmp_dir, "test_cli_config.motifscanrc")
    config = Config(config_file)
    config.set_motif_path("motif_set", "/path/to/motif")
    config.write()

    args = parser.parse_args(["config", "--get-motif", "motif_set"])
    run(args=args, config_file=config_file)
    captured = capsys.readouterr()
    assert captured.out == "/path/to/motif\n"

    with pytest.raises(SystemExit):
        args = parser.parse_args(["config", "--get-motif", "motif_set1"])
        run(args=args, config_file=config_file)


def test_cli_config_rm_motif(tmp_dir):
    config_file = os.path.join(tmp_dir, "test_cli_config.motifscanrc")
    config = Config(config_file)
    config.set_motif_path("motif_set", "/path/to/motif")
    config.write()

    args = parser.parse_args(["config", "--rm-motif", "motif_set"])
    run(args=args, config_file=config_file)
    config = Config(config_file)
    assert not config.has_motif_set("motif_set")

    with pytest.raises(MotifSetNotFoundError):
        config.get_motif_path("motif_set")
    with pytest.raises(SystemExit):
        run(args=args, config_file=config_file)



import pytest

from motifscan.stats import motif_enrichment
from motifscan.motif.matrix import PositionWeightMatrix


def test_stats_motif_enrichment():
    # n_motifs = 2, n_total_input = 3, n_total_control = 5
    pwm1 = PositionWeightMatrix(values=[[1, 0], [0, 0], [0, 1], [1, 1]],
                                name='pwm1')
    pwm2 = PositionWeightMatrix(values=[[1, 0], [0, 0], [0, 1], [1, 1]],
                                name='pwm2')
    pwm3 = PositionWeightMatrix(values=[[1, 0], [0, 0], [0, 1], [1, 1]],
                                name='pwm3')
    pwms = [pwm1, pwm2, pwm3]
    motif_sites = [[[True], [True, True], []],
                   [[], [], []],
                   [[], [], [True]]]
    motif_sites_control = [[[True], [], [True, True, True], [], [True]],
                           [[], [True], [], [True], []],
                           [[], [], [], [], []]]
    enrichment_results = motif_enrichment(
        pwms=pwms, motif_sites=motif_sites,
        motif_sites_control=motif_sites_control)
    assert len(enrichment_results) == 3
    assert enrichment_results[0].n_input == 2
    assert enrichment_results[0].n_control == 3
    assert enrichment_results[0].fold_change == pytest.approx(2 * 5 / 3 / 3)
    assert enrichment_results[1].n_input == 0
    assert enrichment_results[1].n_control == 2
    assert enrichment_results[1].fold_change == 0
    assert enrichment_results[2].n_input == 1
    assert enrichment_results[2].n_control == 0



"""
motifscan.scanner
-----------------

Module for core motif scanner.
"""

import logging
from collections import namedtuple
from multiprocessing import Pool

from motifscan.motif.score import sliding_motif_score

logger = logging.getLogger(__name__)

MotifSite = namedtuple("MotifSite", ['start', 'score', 'strand'])


class Scanner:
    """
    Motif scanner for a list of genomic regions.

    Parameters
    ----------
    genome : `motifscan.genome.Genome`
        A motifscan genome object to extract sequences from.
    regions : list of `GenomicRegion`
        Specifies the genomic regions for scanning.
    window_size : int, optional
        If `window_size` <= 0, then the whole input region will be scanned.
        Otherwise, only scans the the specified window centered at the summit
        or midpoint of each genomic region.
    strand : {'both', '+', '-'}, optional
        Which strand is used to find motif occurrences, default='both'.
    p_value : str, optional
        Significance level for the motif score cutoff, default='1e-4'.
    remove_dup ：bool, optional
        If True (default), remove adjacent duplicated motif occurrences whose
        distance is less than the motif length.
    n_threads : int, optional
        Number of threads used in parallel, default=1.
    """

    def __init__(self, genome, regions, window_size=0, strand='both',
                 p_value='1e-4', remove_dup=True, n_threads=1):
        if window_size <= 0:
            self.window_size = 0
        else:
            self.window_size = window_size
        self.extend = window_size // 2
        if strand in ['both', '+', '-']:
            self.strand = strand
        else:
            raise ValueError(f"invalid strand option: {strand!r}")
        self.p_value = p_value
        self.remove_dup = remove_dup
        self.n_threads = n_threads
        self.seq_starts = []
        self.seq_ends = []
        self.sequences = []
        self.sequences_rc = None
        self._extract_seq(genome=genome, regions=regions)

    def _extract_seq(self, genome, regions):
        """Extract the sequences on the forward strand and record the
        underlying coordinates.
        """
        logger.debug("Extracting sequences on the forward strand")
        for region in regions:
            if self.window_size <= 0:
                seq_start = region.start
                seq_end = region.end
            else:
                seq_start = max(region.summit - self.extend, 0)
                seq_end = min(region.summit + self.extend,
                              genome.chrom_sizes[region.chrom])
            self.seq_starts.append(seq_start)
            self.seq_ends.append(seq_end)
            self.sequences.append(
                genome.fetch_sequence(region.chrom, seq_start, seq_end))
        if self.strand in ['both', '-']:
            self._rc_seq()

    def _rc_seq(self):
        """Reverse complement the sequences for the reverse strand."""
        logger.debug("Extracting sequences on the reverse strand")
        sequences_rc = []
        table = str.maketrans({'a': 't', 'c': 'g', 'g': 'c', 't': 'a',
                               'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'})
        for sequence in self.sequences:
            sequences_rc.append(sequence.translate(table)[::-1])
        self.sequences_rc = sequences_rc

    def _scan_pwm_by_strand(self, pwm, strand):
        """Scan the specified strand for motif occurrences."""
        if strand == '+':
            sequences = self.sequences
        elif strand == '-':
            sequences = self.sequences_rc
        else:
            raise ValueError(f"invalid strand option: {strand!r}")
        score_cutoff = pwm.cutoffs[self.p_value]
        sliding_scores = sliding_motif_score(
            pwm.matrix.tolist(), pwm.length, pwm.max_raw_score, sequences)
        sites = pick_motif_sites(
            sliding_scores, score_cutoff, self.seq_starts, strand=strand)
        if self.remove_dup:
            sites = deduplicate_motif_sites(sites, pwm.length)
        return sites

    def _scan_pwm(self, pwm):
        """Scan with a PWM and returns the detected motif sites."""
        logger.debug(f"Scanning {pwm.name}")
        if self.strand in ['both', '+']:
            sites_fwd = self._scan_pwm_by_strand(pwm=pwm, strand='+')
        if self.strand in ['both', '-']:
            sites_rev = self._scan_pwm_by_strand(pwm=pwm, strand='-')
        if self.strand == '+':
            sites = sites_fwd
        elif self.strand == '-':
            sites = sites_rev
        else:
            sites = [fwd + rev for fwd, rev in zip(sites_fwd, sites_rev)]
        logger.debug(f"Scanned {pwm.name}")
        return sites

    def scan_motifs(self, pwms):
        """Scan for motif occurrences given the motif PWMs.

        Parameters
        ----------
        pwms : `motifscan.motif.MotifPwms`
            Motif Pwms to be scanned.

        Returns
        -------
        motif_sites : nested list, (n_pwms, n_regions, )
            Identified motif occurrences.
        """
        # check if the corresponding motif score cutoff is set
        for pwm in pwms:
            try:
                pwm.cutoffs[self.p_value]
            except (TypeError, KeyError):
                raise ValueError(
                    f"PWM has no motif score cutoff set for P-value "
                    f"{self.p_value!r}")
        if self.n_threads > 1:
            pool = Pool(processes=self.n_threads)
            motif_sites = pool.map(self._scan_pwm, pwms)
        else:
            motif_sites = [self._scan_pwm(pwm) for pwm in pwms]
        return motif_sites


def pick_motif_sites(sliding_scores, cutoff, seq_starts, strand='+'):
    """Given the motif scores of all positions, pick motif occurrences out."""
    if strand not in ['+', '-']:
        raise ValueError(f"expect '+' or '-' for strand, got {strand!r}")
    sites = []
    for region_idx, scores in enumerate(sliding_scores):
        sites_by_region = []
        if strand == '-':
            scores = reversed(scores)
        for pos_idx, score in enumerate(scores):
            if score >= cutoff:
                start = seq_starts[region_idx] + pos_idx
                sites_by_region.append(MotifSite(start, score, strand))
        sites.append(sites_by_region)
    return sites


def deduplicate_motif_sites(sites, length):
    """Deduplicate adjacent motif occurrences with distance less than the motif
    length and only keep the one with higher motif score.
    """
    for sites_by_region in sites:
        idx = 0
        if len(sites_by_region) > 1:
            while idx + 1 < len(sites_by_region):
                site_curr = sites_by_region[idx]
                site_next = sites_by_region[idx + 1]
                if site_next.start - site_curr.start < length:
                    if site_curr.score >= site_next.score:
                        sites_by_region.pop(idx + 1)
                    else:
                        sites_by_region.pop(idx)
                else:
                    idx += 1
    return sites



"""
motifscan.region.parsers
------------------------

Parsers to read the genomic region files.
"""

import logging

from motifscan.exceptions import RegionFileFormatError

logger = logging.getLogger(__name__)


def is_track_header(line):
    """Returns if the line is a header line used in genome tracks/browsers."""
    line = line.strip()
    if line.startswith('#') or line.startswith('track') or line.startswith(
            'browser'):
        return True
    else:
        return False


def is_comment_header(line):
    """Returns if the line is a comment header line."""
    line = line.strip()
    if line.startswith('#'):
        return True
    else:
        return False


def is_macs_header(line):
    """Returns if the line is a header line used in MACS/MACS2 xls."""
    line = line.strip()
    if line.startswith('#') or line.split('\t')[0] == 'chr':
        return True
    else:
        return False


def is_manorm_header(line):
    """Returns if the line is a header line used in MAnorm xls."""
    line = line.strip()
    if line.startswith('#') or line.split('\t')[0] == 'chr':
        return True
    else:
        return False


class RegionParser:
    """Base class for region file parsers."""

    def __init__(self, format):
        self.format = format

    @staticmethod
    def _is_header(line):
        """Abstract method to check if a line is a header line."""
        raise NotImplementedError

    @staticmethod
    def _parse_line(line):
        """Abstract method to parse a line."""
        raise NotImplementedError

    def parse(self, path):
        """Read genomic regions from the given file."""
        with open(path, 'r') as fin:
            line_num = 0
            expect_header = True
            for line in fin:
                line_num += 1
                line = line.strip()
                if not line:  # skip empty lines
                    continue
                if expect_header:
                    if self._is_header(line):
                        logger.debug(
                            f"Detected header at line {line_num}: {line!r}")
                        continue
                    else:
                        expect_header = False
                try:
                    yield self._parse_line(line)
                except (IndexError, ValueError, TypeError):
                    raise RegionFileFormatError(format=self.format,
                                                line_num=line_num, line=line)


class BedRegionParser(RegionParser):
    """Region parser for the BED format."""

    def __init__(self):
        super().__init__('BED')

    @staticmethod
    def _is_header(line):
        return is_track_header(line)

    @staticmethod
    def _parse_line(line):
        fields = line.strip().split('\t')
        chrom = fields[0]
        start = int(fields[1])
        end = int(fields[2])
        try:
            score = float(fields[4])
        except (TypeError, ValueError, IndexError):
            score = None
        summit = None
        return chrom, start, end, summit, score


class Bed3SummitRegionParser(RegionParser):
    """Region parser for the BED3-summit format."""

    def __init__(self):
        super().__init__(format='BED3-summit')

    @staticmethod
    def _is_header(line):
        return is_comment_header(line)

    @staticmethod
    def _parse_line(line):
        fields = line.strip().split('\t')
        chrom = fields[0]
        start = int(fields[1])
        end = int(fields[2])
        summit = int(fields[3])
        score = None
        return chrom, start, end, summit, score


class MacsRegionParser(RegionParser):
    """Region parser for the MACS-xls format."""

    def __init__(self):
        super().__init__(format='MACS-xls')

    @staticmethod
    def _is_header(line):
        return is_macs_header(line)

    @staticmethod
    def _parse_line(line):
        fields = line.strip().split('\t')
        chrom = fields[0]
        start = int(fields[1]) - 1  # coordinates are 1-based in MACS xls
        end = int(fields[2])
        summit = int(fields[4]) + start  # relative summit pos for MACS1
        score = float(fields[6])
        return chrom, start, end, summit, score


class Macs2RegionParser(RegionParser):
    """Region parser for the MACS2-xls format."""

    def __init__(self):
        super().__init__(format='MACS2-xls')

    @staticmethod
    def _is_header(line):
        return is_macs_header(line)

    @staticmethod
    def _parse_line(line):
        fields = line.strip().split('\t')
        chrom = fields[0]
        start = int(fields[1]) - 1  # coordinates are 1-based in MACS2 xls
        end = int(fields[2])
        summit = int(fields[4]) - 1  # absolute summit pos for MACS2
        score = float(fields[6])
        return chrom, start, end, summit, score


class NarrowPeakRegionParser(RegionParser):
    """Region parser for the NarrowPeak format."""

    def __init__(self):
        super().__init__(format='NarrowPeak')

    @staticmethod
    def _is_header(line):
        return is_track_header(line)

    @staticmethod
    def _parse_line(line):
        fields = line.strip().split('\t')
        chrom = fields[0]
        start = int(fields[1])
        end = int(fields[2])
        score = float(fields[4])
        # https://genome.ucsc.edu/FAQ/FAQformat.html#format12
        summit = int(fields[9])
        if summit == -1:
            summit = None
        else:
            summit = start + summit
        return chrom, start, end, summit, score


class BroadPeakRegionParser(RegionParser):
    """Region parser for the BroadPeak format."""

    def __init__(self):
        super().__init__(format='BroadPeak')

    @staticmethod
    def _is_header(line):
        return is_track_header(line)

    @staticmethod
    def _parse_line(line):
        fields = line.strip().split('\t')
        chrom = fields[0]
        start = int(fields[1])
        end = int(fields[2])
        score = float(fields[4])
        # https://genome.ucsc.edu/FAQ/FAQformat.html#format13
        summit = None
        return chrom, start, end, summit, score


class ManormRegionParser(RegionParser):
    def __init__(self):
        super().__init__(format='MAnorm-xls')

    @staticmethod
    def _is_header(line):
        return is_manorm_header(line)

    @staticmethod
    def _parse_line(line):
        fields = line.strip().split('\t')
        chrom = fields[0]
        start = int(fields[1]) - 1
        end = int(fields[2])
        summit = int(fields[3]) - 1
        score = float(fields[4])  # M-value here
        return chrom, start, end, summit, score


def get_region_parser(format):
    """Get proper region parser for the given format.

    Parameters
    ----------
    format : str
        File format (case-insensitive).

    Returns
    -------
        Corresponding region parser.
    """
    format = format.lower()
    if format == 'bed':
        return BedRegionParser
    elif format == 'bed3-summit':
        return Bed3SummitRegionParser
    elif format == 'macs':
        return MacsRegionParser
    elif format == 'macs2':
        return Macs2RegionParser
    elif format == 'narrowpeak':
        return NarrowPeakRegionParser
    elif format == 'broadpeak':
        return BroadPeakRegionParser
    elif format == 'manorm':
        return ManormRegionParser
    else:
        raise ValueError(f"unknown region file format: {format!r}")



"""
motifscan.logging
-----------------

Logging configurations for MotifScan.
"""

import logging
import sys

# Initialize the default logger with a NullHandler
logger = logging.getLogger('motifscan')
logger.addHandler(logging.NullHandler())


class CleanFormatter(logging.Formatter):
    """Clean logging formatter, omit the level label for `logging.INFO`."""

    def format(self, record):
        if record.levelno != logging.INFO:
            record.msg = f"{record.levelname}: {record.msg}"
        return super().format(record)


def setup_logger(verbose=False):
    """Setup logger with a stderr stream handler."""
    logger.setLevel(logging.DEBUG)
    # clear all handlers
    for handler in logger.handlers:
        logger.removeHandler(handler)

    sh = logging.StreamHandler(stream=sys.stderr)
    if verbose:
        sh.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s %(name)-28s %(lineno)-4d %(levelname)-8s %(message)s",
            datefmt="%Y-%m-%d %H:%M")
    else:
        sh.setLevel(logging.INFO)
        formatter = CleanFormatter()
    sh.setFormatter(formatter)
    logger.addHandler(sh)



import os

from motifscan.region import GenomicRegion
from motifscan.region.utils import overlap_with, subset_by_location, \
    generate_control_regions
from motifscan.genome.annotation import read_gene_annotation


def test_overlap_with():
    assert not overlap_with([], 1, 100)
    intervals = [[1, 5], [3, 8], [10, 12]]
    assert overlap_with(intervals, 1, 3)
    assert overlap_with(intervals, 3, 6)
    assert not overlap_with(intervals, 0, 1)
    assert not overlap_with(intervals, 8, 10)
    assert not overlap_with(intervals, 15, 20)


def test_subset_by_location(genome_root):
    genes = read_gene_annotation(
        os.path.join(genome_root, 'test', 'test_gene_annotation.txt'))
    regions = [GenomicRegion(chrom='chr1', start=9868, end=13868)]
    regions_subset = subset_by_location(regions=regions, genes=genes,
                                        location='promoter')
    assert len(regions_subset) == 1
    regions_subset = subset_by_location(regions=regions, genes=genes,
                                        location='distal')
    assert len(regions_subset) == 0
    regions = [GenomicRegion(chrom='chr1', start=9868, end=10868)]
    regions_subset = subset_by_location(regions=regions, genes=genes,
                                        location='promoter', upstream=1000)
    assert len(regions_subset) == 0


def test_generate_control_regions(genome_root):
    genes = read_gene_annotation(
        os.path.join(genome_root, 'test', 'test_gene_annotation.txt'))
    regions = [GenomicRegion(chrom='chr1', start=9868, end=13868),
               GenomicRegion(chrom='chr1', start=50000, end=51000),
               GenomicRegion(chrom='chr1', start=17200, end=17500)]
    regions_random = generate_control_regions(
        n_random=2, regions=regions, chrom_size={'chr1': 1000000},
        genes=genes)
    assert len(regions_random) == 6
    regions_random = generate_control_regions(
        n_random=2, regions=regions, chrom_size={'chr1': 1000000},
        genes=genes, random_seed=1)
    assert len(regions_random) == 6
    regions_random = generate_control_regions(
        n_random=2, regions=regions, chrom_size={'chr1': 1000000})
    assert len(regions_random) == 6



import pytest
from motifscan.region import GenomicRegion


def test_region_init():
    region = GenomicRegion(chrom='chr1', start=1, end=100, score=10)
    assert region.summit == 50
    assert region.score == 10
    region = GenomicRegion(chrom='chr1', start=1, end=100, summit=51)
    assert region.summit == 51
    with pytest.raises(ValueError):
        GenomicRegion(chrom='chr1', start=1, end=1)
    with pytest.raises(ValueError):
        GenomicRegion(chrom='chr1', start=2, end=1)
    with pytest.raises(ValueError):
        GenomicRegion(chrom='chr1', start=1, end=100, summit=0)
    with pytest.raises(ValueError):
        GenomicRegion(chrom='chr1', start=1, end=100, summit=100)



"""
motifscan.region.utils
----------------------

Module for genomic regions utilities.
"""

import logging
import random

from motifscan.region import GenomicRegion

logger = logging.getLogger(__name__)


def overlap_with(intervals, start, end):
    """Returns whether the interval [start, end) overlaps with the intervals.
    The input intervals have been sorted in advance for binary search.

    Parameters
    ----------
    intervals : list
        Intervals.
    start : int
        Start coordinate.
    end : int
        End coordinate.

    Returns
    -------
    bool
        Overlaps or not.
    """
    if not intervals:
        return False
    left = 0
    right = len(intervals) - 1
    while left <= right:
        mid = (left + right) // 2
        start_ref = intervals[mid][0]
        end_ref = intervals[mid][1]
        if not (end <= start_ref or start >= end_ref):
            return True
        elif start >= end_ref:
            left = mid + 1
        elif end <= start_ref:
            right = mid - 1
    return False


def subset_by_location(regions, genes, location, upstream=2000,
                       downstream=2000):
    """Filter genomic regions by location (promoter or distal).

    Parameters
    ----------
    regions: list
        Genomic regions to be filtered.
    genes : `motifscan.genome.annotation.Genes`
        Gene annotations.
    location : {'promoter', 'distal'}
        Only keep regions located at promoters or distal regions.
    upstream : int, optional
        TSS upstream distance to define promoters.
    downstream : int, optional
        TSS downstream distance to define promoters.

    Returns
    -------
        Filtered genomic regions subset.
    """
    filtered_regions = []
    promoters = {}
    for region in regions:
        if region.chrom not in promoters:
            # get promoter intervals of that chromosome
            promoters[region.chrom] = []
            for gene in genes.fetch(region.chrom):
                promoters[region.chrom].append(
                    gene.promoter(upstream, downstream))
            promoters[region.chrom].sort()
        overlap = overlap_with(promoters[region.chrom], region.start,
                               region.end)
        if not overlap ^ (location == 'promoter'):  # xor
            filtered_regions.append(region)
    return filtered_regions


def generate_control_regions(n_random, regions, chrom_size, genes=None,
                             random_seed=None):
    """Generate random control regions for the given reference regions.
    The length and chromosome distribution for each region are controlled to
    match the reference regions. If `genes` and location is specified,

    Parameters
    ----------
    n_random : int
        How many random control regions is generated for each reference region.
    regions : list
        Reference regions.
    chrom_size : dict
        Chromosome sizes.
    genes : `motifscan.genome.annotation.Genes`
        Gene annotations, controls the distances of regions to nearest genes
        when generating control regions. In this way, the control regions
        generated for a promoter region is also picked from promoters.
        If not specified, ignore annotations and random sampling from the
        whole background.
    random_seed: int, optional
        random_seed: The seed used to set random state.
    """
    if random_seed is not None:
        logger.debug(f"Setting random seed: {random_seed}")
        random.seed(random_seed)
    regions_control = []
    for region in regions:
        length = region.end - region.start
        if genes is None:
            for _ in range(n_random):
                start = random.randint(0, chrom_size[region.chrom] - length)
                regions_control.append(
                    GenomicRegion(chrom=region.chrom, start=start,
                                  end=start + length))
        else:
            genes_chrom = genes.fetch(region.chrom)
            distance = dis_to_nearest_gene(region, genes_chrom)
            n_generated = 0
            while n_generated < n_random:
                if distance is None:
                    # randomize a distance if it's too far away gene TSSs
                    distance = random.randint(10000, 100000)
                gene_random = random.choice(genes_chrom)
                if gene_random.strand == '+':
                    start = gene_random.tss + distance
                else:
                    start = gene_random.tss - distance
                if (start >= 0) and (
                        start + length <= chrom_size[region.chrom]):
                    regions_control.append(
                        GenomicRegion(chrom=region.chrom, start=start,
                                      end=start + length))
                    n_generated += 1
    return regions_control


def dis_to_nearest_gene(region, genes, distance_cutoff=10000):
    """Compute the distance of  a specified region to the nearest gene TSS.

    Parameters
    ----------
    region : GenomicRegion
        GenomicRegion object.
    genes : list
        List of genes on that chromosome.
    distance_cutoff : int
        Maximal distance cutoff.

    Returns
    -------
    int or None
        A signed distance, positive if the region is located at the downstream
        of the nearest gene and vice versa. Or None if the distance is beyond
        the distance cutoff.
    """
    dis_min = distance_cutoff
    target_gene = None
    for gene in genes:
        tmp_dis = region.start - gene.tss
        if abs(tmp_dis) < dis_min:
            dis_min = tmp_dis
            target_gene = gene
    if target_gene is None:
        return None
    else:
        if target_gene.strand == '-':
            # signed value, positive for downstream and negative for upstream
            dis_min = -dis_min
        return dis_min



"""
motifscan.config
----------------

MotifScan configuration handler, this configuration manages the paths of
genome and motif data files in the file system.
"""

import os
from configparser import ConfigParser, Error

from motifscan.exceptions import InvalidConfigFileError, GenomeNotFoundError, \
    MotifSetNotFoundError

user_rc_path = os.path.expanduser("~/.motifscanrc")
user_genome_dir = os.path.expanduser("~/.motifscan/genomes/")
user_motif_dir = os.path.expanduser("~/.motifscan/motifs/")


class Config:
    """Configuration handler for MotifScan."""

    _sections = ['motifscan', 'genome', 'motif']

    def __init__(self, path=None):
        self.path = path or user_rc_path
        self._config = ConfigParser(allow_no_value=False)
        try:
            self._config.read(self.path)
        except Error as e:
            raise InvalidConfigFileError(self.path) from e
        # set default for all sections
        for section in self._sections:
            if not self._config.has_section(section):
                self._config.add_section(section)
        # set the default genome root directory
        if not self._config.has_option('motifscan', 'genome_dir'):
            self.set_genome_dir(user_genome_dir)
        # set the default motif root directory
        if not self._config.has_option('motifscan', 'motif_dir'):
            self.set_motif_dir(user_motif_dir)

    def get_genome_dir(self):
        """Get the genome root directory."""
        return self._config.get('motifscan', 'genome_dir')

    def set_genome_dir(self, path):
        """Set the specified path as the genome root directory."""
        self._config.set('motifscan', 'genome_dir', path)

    def get_motif_dir(self):
        """Get the motif root directory."""
        return self._config.get('motifscan', 'motif_dir')

    def set_motif_dir(self, path):
        """Set the specified path as the motif root directory."""
        self._config.set('motifscan', 'motif_dir', path)

    def list_genome_assemblies(self):
        """List configured (installed) genome assemblies."""
        for name, path in self._config.items('genome'):
            yield name, path

    def has_genome_assembly(self, name):
        """Returns if the specified genome assembly is configured."""
        return self._config.has_option('genome', name)

    def get_genome_path(self, name):
        """Get the genome path of the specified genome assembly."""
        if self._config.has_option('genome', name):
            return self._config.get('genome', name)
        else:
            raise GenomeNotFoundError(name)

    def set_genome_path(self, name, path):
        """Set the genome path for the specified genome assembly."""
        self._config.set('genome', name, path)

    def remove_genome_path(self, name):
        """Remove the specified genome assembly out of the configuration."""
        if self._config.has_option('genome', name):
            return self._config.remove_option('genome', name)
        else:
            raise GenomeNotFoundError(name)

    def list_motif_sets(self):
        """List configured (installed) motif PFMs sets."""
        for name, path in self._config.items('motif'):
            yield name, path

    def has_motif_set(self, name):
        """Returns if the specified motif PFMs set is configured."""
        return self._config.has_option('motif', name)

    def get_motif_path(self, name):
        """Get the motif path of the specified motif PFMs set."""
        if self._config.has_option('motif', name):
            return self._config.get('motif', name)
        else:
            raise MotifSetNotFoundError(name)

    def set_motif_path(self, name, path):
        """Set the motif path for the specified motif PFMs set."""
        self._config.set('motif', name, path)

    def remove_motif_path(self, name):
        """Remove the specified motif PFMs set out of the configuration."""
        if self._config.has_option('motif', name):
            return self._config.remove_option('motif', name)
        else:
            raise MotifSetNotFoundError(name)

    def write(self, path=None):
        """Save the configuration."""
        path = path or self.path
        with open(path, 'w') as f_config:
            self._config.write(f_config)



import os

import pytest

from motifscan.genome.annotation import Gene, Genes, RefGeneTxtParser, \
    read_gene_annotation


def test_gene_init():
    gene = Gene(chrom='chr1', tss=10000, strand='+', name='gene1')
    assert gene.chrom == 'chr1'
    assert gene.tss == 10000
    assert gene.strand == '+'
    assert gene.name == 'gene1'
    assert gene.promoter(upstream=4000, downstream=2000) == [6000, 12000]
    gene = Gene(chrom='chr1', tss=10000, strand='-', name='gene2')
    assert gene.promoter(upstream=4000, downstream=2000) == [8000, 14000]
    with pytest.raises(ValueError):
        Gene(chrom='chr1', tss=10000, strand='.', name='bad')


def test_genes_init(genome_root):
    gene_path = os.path.join(genome_root, 'test', 'test_gene_annotation.txt')
    genes = Genes(gene_path)
    assert genes.path == gene_path
    assert len(genes) == 5
    genes_chr1 = genes.fetch('chr1')
    assert len(genes_chr1) == 4
    assert genes.fetch('chrX') == []


def test_ref_gene_txt_parser(genome_root):
    gene_path = os.path.join(genome_root, 'test', 'test_gene_annotation.txt')
    parser = RefGeneTxtParser(gene_path)
    assert parser.path == gene_path
    genes = list(parser.parse())
    assert len(genes) == 5
    assert genes[0].chrom == 'chr1'
    assert genes[0].tss == 11868
    assert genes[0].strand == '+'
    assert genes[0].name == 'NR_148357'
    assert genes[1].chrom == 'chr1'
    assert genes[1].tss == 11873
    assert genes[1].strand == '+'
    assert genes[1].name == 'NR_046018'
    assert genes[2].chrom == 'chr22'
    assert genes[2].tss == 24666798
    assert genes[2].strand == '+'
    assert genes[2].name == 'NM_015330'
    assert genes[3].chrom == 'chr1'
    assert genes[3].tss == 17436
    assert genes[3].strand == '-'
    assert genes[3].name == 'NR_106918'
    assert genes[4].chrom == 'chr1'
    assert genes[4].tss == 17436
    assert genes[4].strand == '-'
    assert genes[4].name == 'NR_107062'
    with pytest.raises(ValueError):
        gene_path = os.path.join(genome_root, 'bad',
                                 'test_gene_annotation_bad.txt')
        parser = RefGeneTxtParser(gene_path)
        list(parser.parse())


def test_read_gene_annotation(genome_root):
    gene_path = os.path.join(genome_root, 'test', 'test_gene_annotation.txt')
    genes = read_gene_annotation(gene_path)
    assert genes.path == gene_path
    assert len(genes) == 5
    genes_chr1 = genes.fetch('chr1')
    assert len(genes_chr1) == 4
    assert genes.fetch('chrX') == []



"""
motifscan.genome.databases
--------------------------

Get genome assemblies from public databases.
"""

import logging
import os
import sys
from collections import namedtuple
from xml.etree import ElementTree

import requests
from tqdm import tqdm

from motifscan.exceptions import RemoteGenomeNotFoundError, \
    RemoteGenomeFileNotFoundError

logger = logging.getLogger(__name__)

Assembly = namedtuple('Assembly', ['id', 'description'])


class UcscDatabase:
    """UCSC genome database.

    Attributes
    ----------
    name : str
        Database name, `UCSC` here.
    info_page : str
        The DAS info URL of the UCSC database.
    seq_url_fmts : list of str
        URL patterns for the sequence file.
    gene_url_fmts : list of str
        URL patterns for the gene annotation file.
    """

    def __init__(self):
        self.name = 'UCSC'
        self._assemblies = None
        self.info_page = "https://genome.ucsc.edu/cgi-bin/das/dsn"
        url_prefix = "http://hgdownload.soe.ucsc.edu/goldenPath/"
        self.seq_url_fmts = [
            url_prefix + "{0}/bigZips/{1}.fa.gz",
            url_prefix + "{0}/bigZips/chromFa.tar.gz",
            url_prefix + "{0}/bigZips/{1}.chromFa.tar.gz",
            url_prefix + "{0}/bigZips/chromFa.zip",
        ]
        self.gene_url_fmts = [url_prefix + "{0}/database/refGene.txt.gz"]

    @property
    def assemblies(self):
        """Returns available genome assemblies in the UCSC database.

        Returns
        -------
        list of `Assembly`
            Available assemblies in the UCSC database. Items are namedtuple
            objects with `id` and `description` as attributes.
        """
        if self._assemblies is None:
            r = requests.get(self.info_page)
            root = ElementTree.fromstring(r.text)
            assemblies = []
            for child in root:
                if child[0].tag == 'SOURCE' and child[1].tag == 'DESCRIPTION':
                    assemblies.append(Assembly(id=child[0].attrib['id'],
                                               description=child[1].text))
            self._assemblies = assemblies
        return self._assemblies

    def search(self, keyword):
        """Search genome assemblies by matching the given keyword either in
        the id field or the description field.

        Parameters
        ----------
        keyword : str
            The keyword to search for (case-insensitive).

        Yields
        ------
        assembly : `Assembly`
            Assemblies matching the specified `keyword`.
        """
        for assembly in self.assemblies:
            if (keyword.lower() in assembly.id.lower() or
                    keyword.lower() in assembly.description.lower()):
                yield assembly

    def get_sequence_url(self, assembly):
        """Returns the sequence file URL of the specified genome assembly.

        Parameters
        ----------
        assembly : str
            Genome assembly name (case-sensitive).

        Returns
        -------
        url : str
            The URL of the sequence file.

        Raises
        ------
        RemoteGenomeNotFoundError
            If the specified assembly is not in the database.
        RemoteGenomeFileNotFoundError
            If the sequence file does not exists or is not accessible.
        """
        found_assembly = False
        for tmp_assembly in self.assemblies:
            if tmp_assembly.id == assembly:
                found_assembly = True
                for pattern in self.seq_url_fmts:
                    url = pattern.format(assembly, assembly)
                    logger.debug(f"Trying URL: {url}")
                    r = requests.head(url)
                    if r.status_code == requests.codes.ok:
                        logger.debug("Succeed")
                        return url
                    else:
                        logger.debug(
                            f"Failed with status code: {r.status_code}")
        if not found_assembly:
            raise RemoteGenomeNotFoundError(self.name, assembly)
        else:
            raise RemoteGenomeFileNotFoundError(self.name, assembly,
                                                'sequence')

    def get_gene_url(self, assembly):
        """Returns the gene annotation file URL of the specified genome
        assembly.

        Parameters
        ----------
        assembly : str
            Genome assembly name (case-sensitive).

        Returns
        -------
        url : str
            The URL of the gene annotation file.

        Raises
        ------
        RemoteGenomeNotFoundError
            If the specified assembly is not in the database.
        RemoteGenomeFileNotFoundError
            If the gene annotation file does not exists or is not accessible.
        """
        found_assembly = False
        for tmp_assembly in self.assemblies:
            if tmp_assembly.id == assembly:
                found_assembly = True
                for pattern in self.gene_url_fmts:
                    url = pattern.format(assembly)
                    logger.debug(f"Trying url: {url}")
                    r = requests.head(url)
                    if r.status_code == requests.codes.ok:
                        logger.debug("Succeed")
                        return url
                    else:
                        logger.debug(
                            f"Failed with status code: {r.status_code}")
        if not found_assembly:
            raise RemoteGenomeNotFoundError(self.name, assembly)
        else:
            raise RemoteGenomeFileNotFoundError(self.name, assembly,
                                                'annotation')

    @staticmethod
    def _download_ucsc_file(url, download_dir):
        """Download file from `url` and save into `download_dir`."""
        if not os.path.isdir(download_dir):
            os.makedirs(download_dir)
        base_name = os.path.basename(url)
        dst = os.path.join(download_dir, base_name)
        try:
            logger.debug(f"Downloading {url}")
            r = requests.get(url, stream=True)
            r.raise_for_status()
            total_size = int(r.headers.get('Content-Length'))
            with open(dst, 'wb') as f:
                with tqdm(total=total_size, unit='B', unit_scale=True,
                          desc=base_name) as pbar:
                    for chunk in r.iter_content(chunk_size=1024):
                        f.write(chunk)
                        pbar.update(len(chunk))
        except requests.HTTPError as e:
            logger.error(f"Failed to download due to an HTTPError: {e}")
            sys.exit(1)
        return dst

    def download_sequence(self, assembly, download_dir):
        """Download the sequence file of the specified genome assembly.

        Parameters
        ----------
        assembly : str
            Genome assembly name (case-sensitive).
        download_dir : str
            Directory to write the downloaded file.

        Returns
        -------
        dst : str
            Path of saved sequence file.
        """
        url = self.get_sequence_url(assembly)
        logger.info("Downloading the sequence file")
        dst = self._download_ucsc_file(url, download_dir)
        return dst

    def download_gene(self, assembly, download_dir):
        """Download the gene annotation file of the specified genome assembly.

        Parameters
        ----------
        assembly : str
            Genome assembly name (case-sensitive).
        download_dir : str
            Directory to write the downloaded file.

        Returns
        -------
        dst : str
            Path of saved gene annotation file.
        """
        url = self.get_gene_url(assembly)
        logger.info("Downloading the gene annotation file")
        dst = self._download_ucsc_file(url, download_dir)
        return dst



"""
motifscan.genome
----------------

This module contains classes and functions implemented for genome assemblies.
"""

import logging
import os

import numpy as np
import pysam

from motifscan.config import Config
from motifscan.exceptions import GenomeFileNotFoundError, BackgroundFormatError
from motifscan.genome.annotation import read_gene_annotation

logger = logging.getLogger(__name__)

bases = 'ACGT'

fasta_path_fmt = os.path.join("{0}", "{1}.fa")
bg_freq_path_fmt = os.path.join("{0}", "{1}_bg_freq.txt")
gene_path_fmt = os.path.join("{0}", "{1}_gene_annotation.txt")


class Genome:
    """Load a pre-installed genome assembly.

    This class handles basic genome properties like chromosome sizes,
    nucleotide frequencies and other advanced functions (fetching sequences,
    sampling regions etc.).

    Parameters
    ----------
    name : str
        The name of a pre-installed genome assembly to be loaded.
    path : str, optional
        Directory where the genome data files are located. If not specified,
        the path will be retrieved from the config file.

    Attributes
    ----------
    name : str
        The name of the genome assembly.
    path : str
        Directory where the genome data files are located.
    fa : `pysam.FastaFile`
        Fasta file handler.
    bg_freq : dict of {str: float}
        Nucleotide frequencies of whole genome background.
    genes : `motifscan.genome.annotation.Genes` or None
        Annotated genes, or None if the gene annotation file is not available.

    Raises
    ------
    GenomeFileNotFoundError
        If required genome data files are missing or broken.
    """

    def __init__(self, name, path=None):
        logger.info(f"Loading genome {name!r}")
        self.name = name
        self.path = path or Config().get_genome_path(self.name)
        self._fasta_path = fasta_path_fmt.format(self.path, self.name)
        self._bg_freq_path = bg_freq_path_fmt.format(self.path, self.name)
        self._gene_path = gene_path_fmt.format(self.path, self.name)
        if os.path.isfile(self._fasta_path):
            self.fa = pysam.FastaFile(self._fasta_path)
        else:
            raise GenomeFileNotFoundError(self.name, 'sequence')
        if os.path.isfile(self._bg_freq_path):
            self.bg_freq = read_bg_freq(self._bg_freq_path)
        else:
            raise GenomeFileNotFoundError(self.name, 'background frequency')
        if os.path.isfile(self._gene_path):
            self.genes = read_gene_annotation(self._gene_path)
        else:
            logger.warning("No gene annotation file found")
            self.genes = None

        self._chroms = None
        self._chrom_sizes = None

    def close(self):
        """Close the fasta handler."""
        self.fa.close()

    @property
    def chroms(self):
        """Returns sorted chromosome names of the genome assembly.

        Returns
        -------
        list of str
            Chromosome names (sorted) of the genome assembly.
        """
        if self._chroms is None:
            self._chroms = sorted(self.fa.references)
        return self._chroms

    @property
    def chrom_sizes(self):
        """Returns the sizes of chromosomes.

        Returns
        -------
        dict of {str: int}
            A dict object with key as the chromosome name and value as the
            size of that chromosome.
        """
        if self._chrom_sizes is None:
            self._chrom_sizes = {chrom: self.fa.get_reference_length(chrom) for
                                 chrom in self.chroms}
        return self._chrom_sizes

    def fetch_sequence(self, chrom, start, end):
        """Returns the sequence of given genomic region (0-based coordinates).
        Wrapper of `pysam.FastaFile.fetch`.

        Parameters
        ----------
        chrom : str
            Chromosome name.
        start : int
            Start position.
        end : int
            End position.

        Returns
        -------
        str
            The sequence of the specified region.
        """
        return self.fa.fetch(chrom, start, end)

    def random_sequences(self, n_times, length, random_seed=None):
        """Random sampling sequences from whole genome. The weight for each
        chromosome to be selected equals to the proportion of its chrom size.
        If the fraction of 'N' bases >= 0.3 in a random sampled sequence,
        the sequence will be discarded and re-sampled.

        Parameters
        ----------
        n_times: int
            Sampling times.
        length: int
            The length of sequences to generate.
        random_seed: int, optional
            random_seed: The seed used to set random state.

        Yields
        ------
        str
            Random sampled sequence of length `length`.
        """
        if random_seed is not None:
            logger.debug(f"Setting random seed: {random_seed}")
            np.random.seed(random_seed)
        chrom_sizes_sum = sum(self.chrom_sizes.values())
        chrom_weight = [self.chrom_sizes[chrom] / chrom_sizes_sum for chrom in
                        self.chroms]
        random_chroms = np.random.choice(self.chroms, size=n_times,
                                         p=chrom_weight)
        max_n = 0.3 * length
        n_seq = 0
        n_loop = 0
        while n_seq < n_times:
            chrom = random_chroms[n_loop % n_times]
            start = np.random.randint(self.chrom_sizes[chrom] - length)
            seq = self.fetch_sequence(chrom, start, start + length)
            if seq.count('N') + seq.count('n') < max_n:
                yield seq
                n_seq += 1
            n_loop += 1


def cal_bg_freq(path, skip_non_autosomes=True):
    """Calculate the nucleotide frequencies of whole genome background.

    Parameters
    ----------
    path : str
        A fasta file which contains the sequences of whole genome background.
    skip_non_autosomes : bool, optional
        If True (default), only count bases on autosomes. Sex chromosomes
        (chrX, chrY), mitochondria DNA (chrM) and other unconfirmed sequences
        (chrUn_*, *_alt*, *_hap*, *_random*) are excluded.

    Returns
    -------
    bg_freq : dict of {str: float}
        A dict object which stores the nucleotide frequencies of A, C, G, T.
    """
    logger.debug(f"Calculating nucleotide frequencies: {path}")
    bg_count = {base: 0 for base in bases}
    keywords_to_skip = ['chrX', 'chrY', 'chrM', 'chrUn_',
                        '_random', '_hap', '_alt']

    fa = pysam.FastaFile(path)
    for chrom in fa.references:
        if skip_non_autosomes:
            skip = False
            for keyword in keywords_to_skip:
                if keyword in chrom:
                    skip = True
                    break
            if skip:
                logger.debug(f"Skipped: {chrom}")
                continue
        logger.debug(f"Processing: {chrom}")
        sequence = fa.fetch(chrom, 0, fa.get_reference_length(chrom)).upper()
        for base in bases:
            bg_count[base] += sequence.count(base)
    fa.close()

    total_count = sum(bg_count.values())
    bg_freq = {base: round(bg_count[base] / total_count, 5) for base in bases}
    return bg_freq


def write_bg_freq(path, bg_freq):
    """Write the nucleotide frequencies of whole genome background.

    Parameters
    ----------
    path : str
        The file path to write the nucleotide frequencies.
    bg_freq: dict of {str: float}
        A dict object which stores the nucleotide frequencies of A, C, G, T.
    """
    logger.debug(f"Writing nucleotide frequencies to {path}")
    with open(path, 'w') as f_out:
        for base in bases:
            f_out.write(f"{base}\t{bg_freq[base]}\n")


def read_bg_freq(path):
    """Read the nucleotide frequencies of whole genome background.

    Parameters
    ----------
    path : str
        The file path to read the nucleotide frequencies.

    Returns
    -------
    bg_freq : dict of {str: float}
        A dict object which stores the nucleotide frequencies of A, C, G, T.

    Raises
    ------
    BackgroundFormatError
        If the frequency file not strictly follows the format.
    """
    logger.debug(f"Reading nucleotide frequencies from {path}")
    bg_freq = {}
    with open(path, 'r') as f_in:
        for idx, expected in enumerate(bases):
            line = f_in.readline().strip()
            base, freq = line.split('\t')
            if base != expected:
                raise BackgroundFormatError(idx + 1, line)
            try:
                bg_freq[base] = float(freq)
            except (ValueError, TypeError):
                raise BackgroundFormatError(idx + 1, line)
    return bg_freq



import os
import pytest

from motifscan.genome import Genome
from motifscan.region import GenomicRegion
from motifscan.scanner import Scanner
from motifscan.motif import MotifPwms
from motifscan.motif.matrix import PositionWeightMatrix


def test_scanner_init(genome_root):
    genome_path = os.path.join(genome_root, 'test')
    genome = Genome(name='test', path=genome_path)
    regions = [GenomicRegion(chrom='chr1', start=2, end=5)]
    scanner = Scanner(genome=genome, regions=regions, window_size=0)
    assert scanner.window_size == 0
    assert scanner.sequences == ['TtC']
    assert scanner.sequences_rc == ['GaA']
    assert scanner.seq_starts == [2]
    assert scanner.seq_ends == [5]
    scanner = Scanner(genome=genome, regions=regions, window_size=4,
                      strand='+')
    assert scanner.sequences_rc is None
    assert scanner.sequences == ['aTtC']
    assert scanner.seq_starts == [1]
    assert scanner.seq_ends == [5]
    with pytest.raises(ValueError):
        Scanner(genome=genome, regions=regions, window_size=0, strand='*')


def test_scan_motif(genome_root):
    genome_path = os.path.join(genome_root, 'test')
    genome = Genome(name='test', path=genome_path)
    regions = [GenomicRegion(chrom='chr1', start=2, end=5)]
    pwms = MotifPwms()
    pwm = PositionWeightMatrix([[1, 0], [0, 1], [0, 0], [1, 0]],
                               cutoffs={'1e-3': 0.5, '1e-4': 1})
    pwms.append(pwm)
    scanner = Scanner(genome=genome, regions=regions, window_size=4,
                      p_value='1e-4')
    sites = scanner.scan_motifs(pwms)
    print(sites)
    assert len(sites[0]) == 1
    assert len(sites[0][0]) == 1
    scanner = Scanner(genome=genome, regions=regions, window_size=4,
                      p_value='1e-3')
    sites = scanner.scan_motifs(pwms)
    print(sites)
    assert len(sites[0][0]) == 3
    scanner = Scanner(genome=genome, regions=regions, window_size=4,
                      p_value='1e-2')
    with pytest.raises(ValueError):
        scanner.scan_motifs(pwms)
    scanner = Scanner(genome=genome, regions=regions, window_size=4,
                      p_value='1e-3', remove_dup=False)
    sites = scanner.scan_motifs(pwms)
    assert len(sites[0][0]) == 5



import logging

from motifscan.logging import logger, setup_logger, CleanFormatter


def test_clean_formatter():
    formatter = CleanFormatter()
    record = logging.makeLogRecord(
        dict(levelname="INFO", levelno=logging.INFO, msg='message info'))
    assert formatter.format(record) == "message info"
    record = logging.makeLogRecord(
        dict(levelname="DEBUG", levelno=logging.DEBUG, msg='message debug'))
    assert formatter.format(record) == "DEBUG: message debug"


def test_setup_logger():
    assert len(logger.handlers) == 1
    assert isinstance(logger.handlers[0], logging.NullHandler)
    setup_logger(verbose=False)
    assert logger.level == logging.DEBUG
    assert len(logger.handlers) == 1
    assert isinstance(logger.handlers[0], logging.StreamHandler)
    assert logger.handlers[0].level == logging.INFO
    setup_logger(verbose=True)
    assert logger.level == logging.DEBUG
    assert len(logger.handlers) == 1
    assert isinstance(logger.handlers[0], logging.StreamHandler)
    assert logger.handlers[0].level == logging.DEBUG



"""
motifscan.genome.annotation
---------------------------

Module for gene annotations.
"""

import logging
from collections import defaultdict

logger = logging.getLogger(__name__)


class Gene:
    """Class for individual gene (transcript)."""

    def __init__(self, chrom, tss, strand, name=None):
        self.chrom = chrom
        self.tss = int(tss)
        if strand not in ['+', '-']:
            raise ValueError(f"invalid strand option: {strand!r}")
        self.strand = strand
        self.name = name

    def promoter(self, upstream=2000, downstream=2000):
        if self.strand == '+':
            return [self.tss - upstream, self.tss + downstream]
        else:
            return [self.tss - downstream, self.tss + upstream]


class Genes:
    """Class for a set of genes."""

    def __init__(self, path):
        self.path = path
        self._genes = defaultdict(list)
        self.read_genes()

    def __len__(self):
        return sum(len(self._genes[chrom]) for chrom in self._genes)

    def fetch(self, chrom):
        if chrom in self._genes:
            return self._genes[chrom]
        else:
            return []

    def read_genes(self):
        logger.debug(f"Loading genes from {self.path}")
        parser = RefGeneTxtParser(self.path)
        for gene in parser.parse():
            self._genes[gene.chrom].append(gene)
        logger.debug(f"Loaded {len(self)} genes")


class RefGeneTxtParser:
    """Parser for refGene in txt format."""

    def __init__(self, path):
        self.path = path

    def parse(self):
        with open(self.path, 'r') as fin:
            for line in fin:
                line = line.strip()
                fields = line.split()
                name = fields[1]
                chrom = fields[2]
                strand = fields[3]
                if strand == '+':
                    tss = int(fields[4])
                elif strand == '-':
                    tss = int(fields[5])
                else:
                    raise ValueError(
                        f"Invalid strand {strand!r} detected at line: {line}")
                yield Gene(chrom=chrom, tss=tss, strand=strand, name=name)


def read_gene_annotation(path):
    return Genes(path)



"""
motifscan.cli.genome
--------------------

Command line interface for the 'genome' subcommand.
"""

import logging
import os
import shutil
import sys

from motifscan.config import Config
from motifscan.exceptions import RemoteGenomeNotFoundError, \
    GenomeNotFoundError
from motifscan.genome import fasta_path_fmt, bg_freq_path_fmt, gene_path_fmt, \
    cal_bg_freq, write_bg_freq
from motifscan.genome.databases import UcscDatabase
from motifscan.io.utils import copy_file, merge_files, merge_extracted_files

logger = logging.getLogger(__name__)


def run(args, config_file=None):
    if args.list:
        config = Config(config_file)
        for name, _ in config.list_genome_assemblies():
            print(name)
        return
    if args.list_remote:
        database = UcscDatabase()
        for assembly in database.assemblies:
            print(f"{assembly.id:12}\t{database.name}\t{assembly.description}")
        return
    if args.search:
        database = UcscDatabase()
        found = False
        for assembly in database.search(args.search):
            found = True
            print(f"{assembly.id:12}\t{database.name}\t{assembly.description}")
        if not found:
            logger.info(f"No match found for {args.search!r}")
        return
    if args.install:
        install_genome(args, config_file)
        return
    if args.uninstall:
        uninstall_genome(args, config_file)
        return


def install_genome(args, config_file=None):
    config = Config(config_file)
    if config.has_genome_assembly(args.name):
        logger.error(f"Genome assembly {args.name!r} already exists!")
        sys.exit(1)

    genome_dir = os.path.abspath(
        args.output_dir or os.path.join(config.get_genome_dir(), args.name))
    logger.info(f"Installing genome assembly {args.name!r} into {genome_dir}")
    if not os.path.isdir(genome_dir):
        os.makedirs(genome_dir)
    if os.listdir(genome_dir):
        logger.error("Directory not empty! Please specify another directory "
                     "or delete files under it.")
        sys.exit(1)

    fasta_path = fasta_path_fmt.format(genome_dir, args.name)
    bg_freq_path = bg_freq_path_fmt.format(genome_dir, args.name)
    gene_path = gene_path_fmt.format(genome_dir, args.name)

    if args.remote:
        download_dir = os.path.join(genome_dir, 'downloads')
        try:
            db = UcscDatabase()
            dst_fasta = db.download_sequence(args.remote, download_dir)
            logger.debug(f"Extracting the sequence file to {fasta_path}")
            merge_extracted_files(dst_fasta, fasta_path)
            dst_gene = db.download_gene(args.remote, download_dir)
            logger.debug(f"Extracting the gene annotation file to {gene_path}")
            merge_extracted_files(dst_gene, gene_path)
            if args.clean:
                logger.debug(f"Removing the download directory {download_dir}")
                shutil.rmtree(download_dir)
        except RemoteGenomeNotFoundError as e:
            logger.error(e)
            sys.exit(1)
    else:
        logger.info("Copying the sequence file(s)")
        merge_files(args.fasta_files, fasta_path)
        logger.info("Copying the gene annotation file")
        copy_file(args.gene_file, gene_path)

    logger.info("Calculating nucleotide frequencies of the genome background")
    bg_freq = cal_bg_freq(fasta_path)
    logger.info("Writing nucleotide frequencies")
    write_bg_freq(bg_freq_path, bg_freq)

    logger.info("Updating the config file")
    config.set_genome_path(args.name, genome_dir)
    config.write()
    logger.info("Successfully installed!")


def uninstall_genome(args, config_file=None):
    try:
        config = Config(config_file)
        path = config.get_genome_path(args.uninstall)
    except GenomeNotFoundError as e:
        logger.error(e)
        sys.exit(1)

    logger.info(f"Uninstalling genome assembly {args.uninstall!r}")
    if os.path.isdir(path):
        logger.info(f"Removing files under {path}")
        try:
            shutil.rmtree(path)
        except Exception as e:
            logger.error(f"Failed to remove the genome directory: {e}")
            sys.exit(1)

    logger.info("Updating the config file")
    config.remove_genome_path(args.uninstall)
    config.write()
    logger.info("Successfully uninstalled!")
